################################################################################
#   Gene prediction pipeline 
#
#   $Id: gff2gff.py 2868 2010-03-03 10:19:52Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
gff2gff.py - manipulate gff files
=================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This scripts reads a :term:`gff` formatted file, operates
on it and outputs the new intervals in :term:`gff` format.

Extension options:

``--extend``
   extend existing features

``--add-up-flank/--add-down-flank``
   add an upstream/downstearm flanking segment to first/last exon of a group.

Segment transformations:

``--crop``
   crop features according to features in a separate gff file.

``--crop-unique``
   remove non-unique features from gff file.

``--merge-features``
   merge consecutive features.

``--join-features``
   group consecutive features

Filtering options:

``--filter-range``
   extract features overlapping a chromosomal range.

``--sanitize``
   reconcile chromosome names between ENSEMBL/UCSC or with an indexed
   genomic fasta file (see :doc:`index_fasta`). Raises an exception if
   an unknown contig is found (unless ``--skip-missing`` is set).

``--remove-contigs``
   remove contig matching to a series of regular expressions.



Usage
-----

Make sure that a :term:`gff` formatted file contains only features on placed chromosomes::

   cat in.gff 
   | gff2gff.py --sanitize=genome --genome-file=hg19 --skip-missing 
   | gff2gff.py --remove-contigs="chrUn,_random" > gff.out

Type::

   python gff2gff.py --help

for command line help.

Command line options
--------------------

'''

import sys
import string
import re
import optparse


import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.GTF as GTF
import CGAT.AGP as AGP
import CGAT.Genomics as Genomics
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Intervals as Intervals
import collections
import numpy
import bx.intervals.io
import bx.intervals.intersection

def combineGFF( gffs, options, merge = True ):
    """join intervals in gff file.

    Note: strandedness is ignored
    """

    if merge:
        min_distance, max_distance, min_features, max_features = map(int, options.merge_features.split(",") )
    else:
        min_distance, max_distance, min_features, max_features = map(int, options.join_features.split(",") )

    E.info( "joining features: min distance=%i, max_distance=%i, at least %i and at most %i features." %
            (min_distance, max_distance, min_features, max_features) )

    def iterate_chunks( gffs ):

        last = gffs.next()
        to_join = [last]

        for gff in gffs:
            d = gff.start - last.end
            if gff.contig == last.contig:
                assert gff.start >= last.start, "input file should be sorted by contig and position: d=%i:\n%s\n%s\n" % (d, last, gff)
            
            if gff.contig != last.contig or \
                    (max_distance and d > max_distance) or \
                    (min_distance and d < min_distance) or \
                    (max_features and len(to_join) >= max_features):
                
                if min_features or len(to_join) >= min_features: 
                    yield to_join
                to_join = []

            last = gff
            to_join.append( gff )
            
        if len(to_join) >= min_features: yield to_join
        raise StopIteration

    id = 1
    ninput, noutput, nfeatures= 0, 0, 0

    if merge:
        for to_join in iterate_chunks(gffs):

            ninput += 1
            y = GTF.Entry()
            t = options.output_format % id
            y.fromGTF( to_join[0], t, t )
            y.start = to_join[0].start
            y.end = to_join[-1].end
            
            options.stdout.write( "%s\n" % str(y ) )
            nfeatures += 1

            noutput += 1
            id += 1
    else:

        for to_join in iterate_chunks(gffs):

            ninput += 1
            for x in to_join:
                y = GTF.Entry()
                t = options.output_format % id
                y.fromGTF( x, t, t )
                options.stdout.write( "%s\n" % str(y ) )
                nfeatures += 1

            noutput += 1
            id += 1

    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, nfeatures=%i\n" % (ninput, noutput, nfeatures) )
        
def cropGFFUnique( gffs, options ):
    """crop intervals in gff file.

    only unique regions are kept. This method ignores the feature field.

    If options.ignore_strand is set, strand information for cropping
    is ignored.
    """

    # read regions to crop with and convert intervals to intersectors
    E.info( "reading gff for cropping: started." )
    gffs = list(gffs)
    if len(gffs) == 0: return

    outf = options.stdout

    def _cmp_without_strand( this, last):
        return this.contig != last.contig

    def _cmp_with_strand( this, last):
        return this.contig != last.contig or this.strand != last.strand

    if options.ignore_strand:
        gffs.sort( key = lambda x: (x.contig, x.start ) )
        comp = _cmp_without_strand
    else:
        gffs.sort( key = lambda x: (x.contig, x.strand, x.start ) )
        comp = _cmp_with_strand

    last = gffs[0]
    c = E.Counter()

    c.input = len(gffs)

    for this in gffs[1:]:
        if comp( this, last) or last.end <= this.start:
            # no overlap
            if last.start < last.end: 
                c.output += 1
                outf.write( "%s\n" % str(last) )
            last = this
            
        elif this.end <= last.start:
            # this ends before last (happens in multiway overlaps)
            # nothing happens
            pass
            
        elif last.end <= this.end:
            # last ends before this
            l = last.end
            last.end = this.start
            this.start = l
            if last.start < last.end: 
                E.info( "overlap" )
                E.info( str(this) )
                E.info( str(last) )
                c.overlaps += 1
                c.output += 1
                outf.write( "%s\n" % str(last) )
            last = this

        elif last.end > this.end:
            # this within last - split last
            l = last.end
            last.end = this.start
            if last.start < last.end: 
                c.output += 1
                c.splits += 1
                outf.write( "%s\n" % str(last) )
            last.start = this.end
            last.end = l

    if last.start < last.end: 
        c.output += 1
        outf.write( "%s\n" % str(last) )

    E.info( "cropping finished: %s" % str(c) )

def cropGFF( gffs, options ):
    """crop intervals in gff file."""

    # read regions to crop with and convert intervals to intersectors
    E.info( "reading gff for cropping: started." )

    other_gffs = GTF.iterator( IOTools.openFile( options.crop, "r") )
    cropper = GTF.readAsIntervals( other_gffs )
    ntotal = 0
    for contig in cropper.keys():
        intersector = bx.intervals.intersection.Intersecter()
        for start, end in cropper[contig]:
            intersector.add_interval( bx.intervals.Interval(start,end) )
            ntotal += 1
        cropper[contig] = intersector

    E.info( "reading gff for cropping: finished." )
    E.info( "reading gff for cropping: %i contigs with %i intervals." % (len(cropper), ntotal ) )

    ninput, noutput, ncropped, ndeleted = 0, 0, 0, 0

    # do the actual cropping
    for gff in gffs:
        
        ninput += 1

        if gff.contig in cropper:
            start, end = gff.start, gff.end
            overlaps = cropper[gff.contig].find(start, end)

            if overlaps:
                l = end - start
                a = numpy.ones( l )
                for i in overlaps:
                    s = max( 0, i.start - start)
                    e = min( l, i.end - start)
                    a[s:e] = 0
                    
                segments = Intervals.fromArray( a )
                
                if len(segments) == 0:
                    ndeleted += 1
                else:
                    ncropped += 1

                for s, e in segments:
                    gff.start, gff.end = s + start, e + start
                    noutput += 1
                    options.stdout.write( "%s\n" % gff )
                    
                continue

        noutput += 1
        options.stdout.write( "%s\n" % gff )
            
    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, ncropped=%i, ndeleted=%i\n" % (ninput, noutput, ncropped, ndeleted) )

##------------------------------------------------------------------------
def main( argv = None ):
    
    if argv == None: argv = sys.argv

    parser = E.OptionParser( version = "%prog version: $Id: gff2gff.py 2868 2010-03-03 10:19:52Z andreas $")

    parser.add_option( "-f", "--forward-coordinates", dest="forward_coordinates", 
                      help="translate to forward coordinates.", action="store_true"  )

    parser.add_option( "--forward-strand", dest="forward_strand", 
                      help="convert to forward strand.", action="store_true"  )

    parser.add_option( "--ignore-strand", dest="ignore_strand", 
                      help="ignore strand information.", action="store_true"  )

    parser.add_option( "--is-gtf", dest="is_gtf", action="store_true",
                       help="input will be treated as gtf [default=%default]." ) 

    parser.add_option( "--add-up-flank", dest="add_up_flank", type="int", 
                      help="add an upstream flanking segment to first exon of a group."  )

    parser.add_option( "--add-down-flank", dest="add_down_flank", type="int", 
                      help="add a downstream flanking segment to last segment of a group."  )

    parser.add_option( "--extend", dest="extend", 
                      help="extend the existing features.", action="store_true"  )

    parser.add_option( "-c", "--contigs", dest="input_filename_contigs", type="string",
                       help="filename with contig lenghts."  )

    parser.add_option( "--filename-agp", dest="input_filename_agp", type="string",
                       help="agp file to map coordinates from contigs to scaffolds." )
    
    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome."  )

    parser.add_option( "--complement-groups", dest="complement_groups", action="store_true",
                       help="""complement groups. Will write introns from exons.""" )

    parser.add_option( "--combine-groups", dest="combine_groups", action="store_true",
                       help="""combine groups.""" )

    parser.add_option( "--filter-range", dest="filter_range", type="string",
                       help="""extract all elements overlapping a range. A range is specified by eithor 'contig:from..to', 'contig:+:from..to', or 'from,to' .""" )

    parser.add_option( "--join-features", dest="join_features", type="string",
                       help="join features into a single transcript. Consecutive features are grouped "
                       " into the same transcript/gene. This metdo expects a string of for numbers ``a,b,c,d`` " 
                       " as input with:"
                       " a,b=minimum/maximum distance between features, "
                       " c,d=minimum,maximum number of features.""" )

    parser.add_option( "--merge-features", dest="merge_features", type="string",
                       help="merge features. Consecutive features are merged into a single feature. "
                       "This method expects a string of four numbers ``a,b,c,d`` as input; "
                       "a,b=minimum/maximum distance between features, "
                       "c,d=minimum,maximum number of features." )

    parser.add_option( "--crop-unique", dest="crop_unique", action="store_true",
                       help = "crop overlapping intervals, keeping only intervals that are unique [default=%default]" )

    parser.add_option( "--crop", dest="crop", type="string",
                       help="""crop features in gff file with features in another file. If a feature falls in the middle of another, two entries will be output.""" )

    parser.add_option( "--sanitize", dest="sanitize", type="choice",
                       choices=( "ucsc", "ensembl", "genome" ),
                       help="sanitize chr names for ucsc or ensembl or use the genome translator [%default]." )

    parser.add_option( "--skip-missing", dest="skip_missing", action="store_true",
                       help="skip entries on missing contigs. Otherwise an exception is raised [%default]." )

    parser.add_option( "--remove-contigs", dest="remove_contigs", type="string", action="store",
                       help="a comma separated list of regular expressions specifying contigs to be removed when runnnig sanitize [%default]." )

    parser.set_defaults(
        forward_coordinates = False,
        forward_strand = False,
        input_filename_contigs = False,
        input_filename_agp = False,
        genome_file = None,
        sanitize = None,
        add_up_flank = None,
        add_down_flank = None,
        extend = False,
        complement_groups=False,
        combine_groups=False,
        crop = None,
        crop_unique = False,
        ignore_strand = False,
        filter_range = None,
        join_features = None,
        merge_features = None,
        output_format = "%06i",
        skip_missing = False,
        remove_contigs = None,
        is_gtf = False,
        )

    (options, args) = E.Start( parser, argv = argv )

    if options.input_filename_contigs:
        contigs = Genomics.ReadContigSizes( IOTools.openFile( options.input_filename_contigs, "r") )

    if options.genome_file:
        genome_fasta = IndexedFasta.IndexedFasta( options.genome_file )
        contigs = genome_fasta.getContigSizes()
    else:
        genome_fasta = None

    if (options.forward_coordinates or options.forward_strand) and not contigs: 
        raise ValueError( "inverting coordinates requires genome file")

    if options.input_filename_agp:
        agp = AGP.AGP()
        agp.readFromFile( IOTools.openFile( options.input_filename_agp, "r" ) )
    else:
        agp = None
        
    if options.is_gtf:
        gffs = GTF.iterator( options.stdin )
    else:
        gffs = GTF.iterator( options.stdin )

    if options.add_up_flank or options.add_down_flank:
        
        if options.is_gtf:
            iterator = GTF.flat_gene_iterator( gffs )
        else:
            iterator = GTF.joined_iterator( gffs )
        
        for chunk in iterator:
            is_positive = Genomics.IsPositiveStrand( chunk[0].strand )
            chunk.sort( lambda x, y: cmp( x.start, y.start) )
            lcontig = contigs[chunk[0].contig]

            if options.extend:
                if options.add_up_flank:
                    if is_positive:
                        chunk[0].start = max( 0, chunk[0].start - options.add_up_flank )
                    else:
                        chunk[-1].end = min( lcontig, chunk[-1].end + options.add_up_flank )
                if options.add_down_flank:
                    if is_positive:
                        chunk[-1].end = min( lcontig, chunk[-1].end + options.add_down_flank )
                    else:
                        chunk[0].start = max( 0, chunk[0].start - options.add_down_flank )
            else:
                if options.add_up_flank:
                    gff = GTF.Entry()
                    if is_positive:
                        gff.copy( chunk[0] )
                        gff.end = gff.start
                        gff.start = max( 0, gff.start - options.add_up_flank )
                        chunk.insert(0, gff)
                    else:
                        gff.copy( chunk[-1] )
                        gff.start = gff.end
                        gff.end = min( lcontig, gff.end + options.add_up_flank )
                        chunk.append( gff )
                    gff.feature = "5-Flank"
                    gff.mMethod = "gff2gff"
                if options.add_down_flank:
                    gff = GTF.Entry()
                    if is_positive:
                        gff.copy( chunk[-1] )
                        gff.start = gff.end
                        gff.end = min( lcontig, gff.end + options.add_up_flank )
                        chunk.append( gff )
                    else:
                        gff.copy( chunk[0] )
                        gff.end = gff.start
                        gff.start = max( 0, gff.start - options.add_up_flank )
                        chunk.insert(0, gff)
                    gff.feature = "3-Flank"
                    gff.mMethod = "gff2gff"

            if not is_positive: chunk.reverse()

            for gff in chunk:
                options.stdout.write( str(gff) + "\n" )

    elif options.complement_groups:
        
        iterator = GTF.joined_iterator( gffs )
        for chunk in iterator:
            chunk.sort()
            x = GTF.Entry()
            x.copy( chunk[0] )
            x.start = x.end
            x.feature = "intron"
            for c in chunk[1:]:
                x.end = c.start
                options.stdout.write( str(x) + "\n" )
                x.start = c.end
            
    elif options.combine_groups:
        
        iterator = GTF.joined_iterator( gffs )

        for chunk in iterator:
            chunk.sort()
            x = GTF.Entry()
            x.copy( chunk[0] )
            x.end = chunk[-1].end
            x.feature = "segment"
            options.stdout.write( str(x) + "\n" )

    elif options.join_features:

        combineGFF( gffs, options, merge = False )

    elif options.merge_features:

        combineGFF( gffs, options, merge = True )

    elif options.crop:
        
        cropGFF( gffs, options )

    elif options.crop_unique:
        
        cropGFFUnique( gffs, options )

    elif options.filter_range:

        contig, strand, interval = None, None, None
        try:
            contig, strand, start, sep, end = re.match( "(\S+):(\S+):(\d+)(\.\.|-)(\d+)", options.filter_range ).groups()
        except AttributeError:
            pass
        
        if not contig:
            try:
                contig, start, sep, end = re.match( "(\S+):(\d+)(\.\.|-)(\d+)", options.filter_range ).groups()
                strand = None
            except AttributeError:
                pass
            
        if not contig:
            try:
                start,end = re.match( "(\d+)(\.\.|\,|\-)(\d+)", options.filter_range).groups()
            except AttributeError:
                raise "can not parse range %s" % options.filter_range
            contig = None
            strand = None
            
        if start:
            interval = (int(start), int(end) )
        else:
            interval = None

        if options.loglevel >= 2:
            options.stdlog.write("# filter: contig=%s, strand=%s, interval=%s\n" % (str(contig), str(strand), str(interval)))
            options.stdlog.flush()

        for gff in GTF.iterator_filtered( gffs, contig = contig, strand = strand, interval = interval ):
            options.stdout.write( str(gff) + "\n" )

    elif options.sanitize:

        def toUCSC(id):
            if not id.startswith("contig") and not id.startswith("chr"):
                id = "chr%s" % id
            return id

        def toEnsembl(id):
            if id.startswith("contig"): return id[len("contig"):]
            if id.startswith("chr"): return id[len("chr"):]
            return id

        if options.sanitize == "genome":
            if genome_fasta == None:
                raise ValueError("please specify --genome-file= when using --sanitize=genome")
            f = genome_fasta.getToken
        elif options.sanitize == "ucsc":
            f = toUCSC
        elif options.sanitize == "ensembl":
            f = toEnsembl

        skipped_contigs = collections.defaultdict( int )
        outofrange_contigs = collections.defaultdict( int )
        filtered_contigs = collections.defaultdict( int )
        
        for gff in gffs:
            try:
                gff.contig = f( gff.contig )
            except KeyError, msg:
                if options.skip_missing:
                    skipped_contigs[gff.contig] += 1
                    continue
                else:
                    raise

            if genome_fasta:
                lcontig = genome_fasta.getLength( gff.contig )
                if lcontig < gff.end:
                    outofrange_contigs[gff.contig] += 1
                    continue

            if options.remove_contigs:
                to_remove = [re.compile(x) for x in options.remove_contigs.split(",")]
                if any([x.match(gff.contig) for x in to_remove]):
                    filtered_contigs[gff.contig] += 1 
                    continue
                
            options.stdout.write( str(gff) + "\n" )

        if skipped_contigs:
            E.info( "skipped %i entries on %i contigs: %s" % (sum(skipped_contigs.values()),
                                                              len(skipped_contigs.keys()),
                                                              str(skipped_contigs) ) )
        if outofrange_contigs:
            E.warn( "skipped %i entries on %i contigs because they are out of range: %s" %\
                    (sum(outofrange_contigs.values()),
                     len(outofrange_contigs.keys()),
                     str(outofrange_contigs) ) )

        if filtered_contigs:
            E.info( "filtered out %i entries on %i contigs: %s" %\
                        (sum(filtered_contigs.values()), 
                         len(filtered_contigs.keys()), 
                         str(filtered_contigs) ) ) 

    else:

        
        for gff in gffs:

            if options.forward_coordinates:
                gff.invert( contigs[gff.contig] )

            if options.forward_strand:
                gff.invert( contigs[gff.contig] )
                gff.strand = "+"

            if agp:
                # note: this works only with forward coordinates
                gff.contig, gff.start, gff.end = agp.mapLocation( gff.contig, gff.start, gff.end )
        
            options.stdout.write( str(gff) + "\n" )

    E.Stop()
    
if __name__ == "__main__":
    sys.exit( main( sys.argv ))



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
WrapperIDR.py - wrap idr analysis
=================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Wrap the IDR analysis workflow.

For more information, see

https://sites.google.com/site/anshulkundaje/projects/idr


Usage
-----

Documentation
-------------

Code
----

This script requires some R functions within the package
called 

functions-all-clayton-12-13.r

These functions have been copied into a script called
WrapperIDR.r in the scripts directory.

'''

import os, sys, re, optparse, tempfile, shutil, subprocess
import collections

import Experiment as E
import IOTools

## for zinba
from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.vectors as rovectors
from rpy2.rinterface import RRuntimeError

def runIDR( options, peakfile1, peakfile2 ):
    '''run IDR analysis.

    This code is taken from the R script
    
    batch-consistency-analysis.r
    '''

    if options.half_width != None:
        R.assign( "half.width", options.half_width )
    else:
        R('''half.width = NULL''') 
    R.assign( "overlap.ratio", options.overlap_ratio )
    R.assign( "is.broadpeak", options.is_broadpeak )
    R.assign( "sig.value", options.signal_value )

    dirname = os.path.dirname(__file__)
    R.source(os.path.join( dirname, "WrapperIDR.r"))

    # read the length of the chromosomes, which will be used to concatenate chr's
    R('''chr.size <- read.table('%s', sep='\t')''' % options.filename_chromosome_table)

    output_prefix = options.output_prefix
    output_uri = output_prefix + "-uri.sav"
    output_em = output_prefix + "-em.sav"
    output_overlapped_peaks = output_prefix + "-overlapped-peaks.txt"
    output_peaks_above_idr = output_prefix + "-npeaks-aboveIDR.txt"

    # process data, summit: the representation of the location of summit
    E.info("loading data" )
    R('''rep1 <- process.narrowpeak('%(peakfile1)s', chr.size, 
                     half.width=half.width, summit="offset", broadpeak=is.broadpeak)''' % locals())
    R('''rep2 <- process.narrowpeak('%(peakfile2)s', chr.size, 
                     half.width=half.width, summit="offset", broadpeak=is.broadpeak)''' % locals())

    E.info( "replicate 1: read %s: %i peaks, %i after filtering" % \
                (peakfile1, 
                 R('''nrow(rep1$data.ori)''')[0], 
                 R('''nrow(rep1$data.cleaned)''')[0]))
    E.info( "replicate 2: read %s: %i peaks, %i after filtering" % \
                (peakfile2, 
                 R('''nrow(rep2$data.ori)''')[0], 
                 R('''nrow(rep2$data.cleaned)''')[0]))

    E.info( "computing correspondence profile (URI)")
        
    R('''uri.output <- compute.pair.uri(rep1$data.cleaned, rep2$data.cleaned, 
                                        sig.value1=sig.value, sig.value2=sig.value, 
                                        overlap.ratio=overlap.ratio)''')
    E.info( "saving correspondence profile to %s" % output_uri )
    R('''save(uri.output, file='%(output_uri)s') ''' % locals())

    E.info( "computing EM procedure for inference")
    R('''em.output <- fit.em(uri.output$data12.enrich, fix.rho2=T)''')
    E.info( "saving EM to %s" % output_em )
    R('''save(em.output, file='%(output_em)s') ''' % locals())

    # write em output into a file
    # cat(paste("EM estimation for the following files\n", peakfile1, "\n", peakfile2, "\n", sep=""))

    options.stdout.write( "em_estimation\n%s\n" % str(R('''em.output$em.fit$para''')))

    # add on 3-29-10
    # output both local idr and IDR
    E.info( "writing overlapped peaks to %s" % output_overlapped_peaks)
    R('''idr.local <- 1-em.output$em.fit$e.z''')
    R('''IDR <- c()''')
    R('''o <- order(idr.local)''')
    R('''IDR[o] <- cumsum(idr.local[o])/c(1:length(o))''')
    R('''
    write.out.data <- data.frame(chr1=em.output$data.pruned$sample1[, "chr"],
                                 start1=em.output$data.pruned$sample1[, "start.ori"],
                                 stop1=em.output$data.pruned$sample1[, "stop.ori"],
                                 sig.value1=em.output$data.pruned$sample1[, "sig.value"],
                                 chr2=em.output$data.pruned$sample2[, "chr"],
                                 start2=em.output$data.pruned$sample2[, "start.ori"],
                                 stop2=em.output$data.pruned$sample2[, "stop.ori"],
                                 sig.value2=em.output$data.pruned$sample2[, "sig.value"],
                                 idr.local=1-em.output$em.fit$e.z, IDR=IDR)
    ''')
    R('''write.table(write.out.data, file='%(output_overlapped_peaks)s')''' % locals())

    # number of peaks passing IDR range (0.01-0.25)
    E.info("computing number of peaks at various thresholds")
    R('''IDR.cutoff <- seq(0.01, 0.25, by=0.01)''')
    R('''idr.o <- order(write.out.data$idr.local)''')
    R('''idr.ordered <- write.out.data$idr.local[idr.o]''')
    R('''IDR.sum <- cumsum(idr.ordered)/c(1:length(idr.ordered))''')
    R('''
    IDR.count <- c()
    n.cutoff <- length(IDR.cutoff)
    for(i in 1:n.cutoff){
        IDR.count[i] <- sum(IDR.sum <= IDR.cutoff[i])
        }
    ''')
    
    # write the number of peaks passing various IDR ranges into a file
    E.info( "writing number of peaks above IDR cutoffs in range [0.01, 0.25] to %s" % output_peaks_above_idr)
    R('''idr.cut <- data.frame( cutoff=IDR.cutoff, count=IDR.count)''')
    R('''write.table(idr.cut, file='%(output_peaks_above_idr)s', quote=F, 
                     row.names=F, col.names=T, sep='\t')''' % locals())

    R('''mar.mean <- get.mar.mean(em.output$em.fit)''')
    options.stdout.write( "marginal mean of two components\n%s\n)" % R('''print(mar.mean)'''))

def plotIDR( output_file, input_prefixes ):
    '''create IDR plots.

    This code is taken from the R script

    batch-consistency-plot.r

    within the IDR package.
    '''

    dirname = os.path.dirname(__file__)
    R.source(os.path.join( dirname, "WrapperIDR.r"))

    R('''df.txt = 10''')    

    R('''uri.list <- list()
         uri.list.match <- list()
         ez.list <- list()
         legend.txt <- c()
         em.output.list <- list()
         uri.output.list <- list()''')

    npair = len(input_prefixes)
    for x, input_prefix in enumerate(input_prefixes):

        R.load( input_prefix + "-uri.sav" )
        R.load( input_prefix + "-em.sav" )
        i = x + 1

        R( '''uri.output.list[[%(i)i]] <- uri.output;
              em.output.list[[%(i)i]] <- em.output;
              # reverse =T for error rate;''' % locals())
        R('''
              ez.list[[%(i)i]] <- get.ez.tt.all(em.output, uri.output.list[[%(i)i]]$data12.enrich$merge1,
                                        uri.output.list[[%(i)i]]$data12.enrich$merge2);''' % locals())
        R('''
              # URI for all peaks
              uri.list[[%(i)i]] <- uri.output$uri.n;

              # URI for matched peaks
              uri.match <- get.uri.matched(em.output$data.pruned, df=df.txt);
              uri.list.match[[%(i)i]] <- uri.match$uri.n;
         ''' % locals() )

        legend = "%(i)i = %(input_prefix)s" % locals()
        R('''
              legend.txt[%(i)i] <- '%(legend)s';
        '''% locals())
        
    R.pdf( output_file )
    R('''par(mfcol=c(2,3), mar=c(5,6,4,2)+0.1)''')
    R('''plot.uri.group(uri.list, NULL, file.name=NULL, c(1:%(npair)i), title.txt="all peaks");
         plot.uri.group(uri.list.match, NULL, file.name=NULL, c(1:%(npair)i), title.txt="matched peaks");
         plot.ez.group(ez.list, plot.dir=NULL, file.name=NULL, legend.txt=c(1:%(npair)i), y.lim=c(0, 0.6));
         plot(0, 1, type="n", xlim=c(0,1), ylim=c(0,1), xlab="", ylab="", xaxt="n", yaxt="n"); 
         legend(0, 1, legend.txt, cex=0.6);''' % locals())
    R["dev.off"]()

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-o", "--output-prefix", dest="output_prefix", type="string",
                      help="output filename prefix [default=%default]."  )

    parser.add_option("-c", "--chromosome-table", dest="filename_chromosome_table", type="string",
                      help="filename with tab separated list of chromosome names [default=%default]."  )

    parser.add_option( "--action", dest="action", type="choice",
                       choices=("plot", "run"),
                       help="action to perform [default=%default]")

    
    parser.set_defaults(
        action = "plot",
        output_prefix = "output",
        half_width = None,
        overlap_ratio = 0,
        is_broadpeak = False,
        signal_value = "signal.value",
        filename_chromosome_table = "genome_table.txt",
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if options.action == "plot":
        plotIDR( options.output_prefix + ".pdf", args)
    elif options.action == "run":
        if len(args) != 2:
            raise ValueError("require exactly two replicates")
        runIDR( options, args[0], args[1])
        
    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )

    



#!/usr/bin/env python

"""
wiggle_extract_ranges_indexed.py - 
===============================================================================

Reads a list of intervals and a wiggle file. Produces a list of 
values similar to wiggle_to_simple.py. 

The difference is that this file allows indexed access to a
set of wiggle files indexed by wiggle_build_index.py

It is assumed that each file `wiggle_fname` has a corresponding `wiggle_fname`.index
file.

WARNING: bz2/bz2t support and file cache support are new and not as well
         tested.

Note: this requires a patched bx/wiggle.py

usage: %prog wiggle_fname1 wiggle_fname2 ... [options] < interval_file
   -v, --version:    Output version
   -s, --src=s:      Use this src for all intervals
   -p, --prefix=p:   Prepend this to each src before lookup
   -S, --strand:     Strand is included as an additional column.
   -C, --usecache:   Use a cache that keeps blocks of the WIGGLE files in memory (requires ~20MB per WIGGLE)
"""

import psyco_full

from bx.cookbook import doc_optparse

import CGAT.Wiggle as Wiggle
import sys
import os.path

def main():
    # Parse Command Line
    options, args = doc_optparse.parse( __doc__ )
    if options.version: return

    try:
        wiggle_files = args
        if options.src: fixed_src = options.src
        else: fixed_src = None
        if options.prefix: prefix = options.prefix
        else: prefix = None
        do_strand = bool( options.strand )
        use_cache = bool( options.usecache )
    except:
        doc_optparse.exit()

    # Open indexed access to wiggles
    index = Wiggle.WiggleMultiIndexedAccess( wiggle_files,
                                             keep_open = True,
                                             use_cache=use_cache )

    for line in sys.stdin:
        strand = "+"
        fields = line.split()
        if fixed_src:
            src, start, end = fixed_src, int( fields[0] ), int( fields[1] )
            if do_strand: strand = fields[2]
        else:
            src, start, end = fields[0], int( fields[1] ), int( fields[2] )
            if do_strand: strand = fields[3]
        if prefix: src = prefix + src
        blocks = index.get( src, start, end )
        for x, values in blocks:
            for v in values:
                print "%s\t%i\t%f" % (src, x, v)
                x += 1
        
if __name__ == "__main__": main()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2cleaned_mali.py - remove split genes from a multiple alignment
===================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Remove split genes from a multiple alignment. Split genes confuse
trees as both parts might not cluster together in the tree. This then
causes problems on tree-reconciliation.

Usage
-----

Example::

   python mali2cleaned_mali.py --help

Type::

   python mali2cleaned_mali.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse

USAGE="""python %s [OPTIONS] < graph.in > graph.out

Version: $Id: mali2cleaned_mali.py 2781 2009-09-10 11:33:14Z andreas $



""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.Mali as Mali
import CGAT.Exons as Exons

import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import networkx

###########################################################
def AddSynonym( rep_transcript, mem_transcript, map_synonyms, map_to_synonyms):

    if rep_transcript not in map_synonyms:
        map_synonyms[rep_transcript] = []

    if mem_transcript in map_synonyms:            

        for s in map_synonyms[mem_transcript]:
            map_to_synonyms[s] = rep_transcript

        map_synonyms[rep_transcript] += map_synonyms[mem_transcript]                    
        del map_synonyms[mem_transcript]

    map_synonyms[rep_transcript].append( mem_transcript )
    map_to_synonyms[mem_transcript] = rep_transcript

###########################################################
def filterJoiningTranscripts( links, options ):
    """cluster links in infile by transcript."""
    
    map_priority = {}
    for x in range( len(options.quality_priority) ):
        map_priority[options.quality_priority[x]] = x
        
    transcripts = {}
    ninput, noutput, nskipped = 0, 0, 0

    ## build matrix
    matrix = {}
    
    for transcript1, transcript2, weight in links:

        weight = float(weight)
        ninput += 1
        
        if options.min_weight != None and weight < options.min_weight:
            nskipped += 1
            continue

        if options.max_weight != None and weight > options.max_weight:
            nskipped += 1
            continue

        species1 = transcript1.split(options.separator)[0]
        species2 = transcript2.split(options.separator)[0]

        if transcript1 > transcript2: transcript1, transcript2 = transcript2, transcript1
        
        if species1 not in transcripts: transcripts[species1] = []
        if species2 not in transcripts: transcripts[species2] = []
        
        transcripts[species1].append( transcript1 )
        transcripts[species2].append( transcript2 )    

        if transcript1 not in matrix: matrix[transcript1] = {}
        if transcript2 not in matrix: matrix[transcript2] = {}        
        
        matrix[transcript1][transcript2] = weight
        matrix[transcript2][transcript1] = weight        

        noutput += 1
        
    ## select each transcripts for each species and go through list
    ## would be more efficient if I had the length

    
    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, nskipped==%i\n" % (ninput, noutput, nskipped))

##-------------------------------------------------------------------------------
def hasGenomicOverlap( key1, key2, exons ):
    """returns true if two transcripts overlap on the genome."""
    ee = exons[key1]
    contig1, strand1, start1, end1 = ee[0].mSbjctToken, ee[0].mSbjctStrand, ee[0].mGenomeFrom, ee[-1].mGenomeTo
    ee = exons[key2]
    contig2, strand2, start2, end2 = ee[0].mSbjctToken, ee[0].mSbjctStrand, ee[0].mGenomeFrom, ee[-1].mGenomeTo

    return contig1 == contig2 and strand1 == strand2 and min(end1,end2) - max(start1, start2) > 0

##-------------------------------------------------------------------------------
def isSynchronous( key1, key2, mali, exons ):
    """check if two transcripts are on synchronous positions.

    Two transcripts are synchroneous, if their position in the
    alignment and on the genomic sequence are both a < b.

    Both front and end are taken into account. Note that they
    can still be overlapping.
    """

    def getStartEnd( s ):
        start, end = 0, len(s)
        x = re.match("([-.]+)", s )
        if x: start = len(x.groups()[0])
        x = re.search("([-.]+)$", s )
        if x: end -= len(x.groups()[0])        
        
        return start, end

    mstart1, mend1 = getStartEnd( mali[key1] )
    mstart2, mend2 = getStartEnd( mali[key2] )    
    
    gstart1, gend1 = exons[key1][0].mGenomeFrom, exons[key1][-1].mGenomeTo
    gstart2, gend2 = exons[key2][0].mGenomeFrom, exons[key2][-1].mGenomeTo    

    return (mstart1 < mstart2 and mend1 < mend2 and gstart1 < gstart2 and gend1 < gend2) or \
           (mstart2 < mstart1 and mend2 < mend1 and gstart2 < gstart1 and gend2 < gend1)

##-------------------------------------------------------------------------------
def getGenomicDistance( key1, key2, exons ):
    """returns the number of nucleotides between two genes.

    return negative value if they are on different contigs/strands or
    are overlapping.
    """
    ee = exons[key1]
    contig1, strand1, start1, end1 = ee[0].mSbjctToken, ee[0].mSbjctStrand, ee[0].mGenomeFrom, ee[-1].mGenomeTo
    ee = exons[key2]
    contig2, strand2, start2, end2 = ee[0].mSbjctToken, ee[0].mSbjctStrand, ee[0].mGenomeFrom, ee[-1].mGenomeTo

    if contig1 == contig2 and strand1 == strand2:
        return -(min(end1,end2) - max(start1, start2))
    else:
        return -1

##-------------------------------------------------------------------------------
def getPercentOverlap( seq1, seq2, gap_chars = (".", "-") ):
    """calculate overlap distance between two sequences."""
    naligned, nunaligned = 0, 0
    for c1, c2 in zip( seq1.upper(), seq2.upper()):
        if c1 in gap_chars or c2 in gap_chars:
            if not (c1 in gap_chars and c2 in gap_chars):
                nunaligned += 1
        else:
            naligned += 1
        
    return 100.0 * float( naligned) / float(nunaligned + naligned)

##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------    
def removeJoiningTranscripts( mali, exons, map_species2transcripts, options ):
    
    ###############################################################
    ###############################################################
    ###############################################################        
    ## identify all joining transcripts
    ###############################################################
    joining_transcripts = set()
    
    for species, transcripts in map_species2transcripts.items():

        transcripts.sort()
        transcripts.reverse()

        for t1 in range(len(transcripts)-1):
            transcript1 = transcripts[t1][1]

            for t2 in range(t1+1, len(transcripts)):
                transcript2 = transcripts[t2][1]
                
                if not hasGenomicOverlap( transcript1, transcript2, exons ):
                    continue
                
                for t3 in range(t2+1, len(transcripts)):
                    transcript3 = transcripts[t3][1]

                    if not hasGenomicOverlap( transcript1, transcript3, exons ):
                        continue

                    if hasGenomicOverlap( transcript2, transcript3, exons ):
                        continue
                    
                    overlap12 = getPercentOverlap( mali[transcript1], mali[transcript2] )
                    overlap13 = getPercentOverlap( mali[transcript1], mali[transcript3] )                    
                    overlap23 = getPercentOverlap( mali[transcript2], mali[transcript3] )

                    if options.loglevel >= 2:
                        options.stdlog.write("# transcript: %s connects %s and %s - mali overlap: %i %i %i\n" % (transcript1, transcript2, transcript3, overlap12, overlap13, overlap23) )
                        joining_transcripts.add( (transcript1, transcript2, "joining transcript") )
                        
    return joining_transcripts

##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------        
def removeSplitTranscripts( mali, exons, map_species2transcripts, options ):
    """remove split transcripts from mulitple alignment.

    This procedure finds two genes of the same species and if they do not overlap
    on the multiple alignment will remove the shorter one. The algorithm works like this

    1. Sort all transcripts per species by length
    2. In descending length:
       Check overlap with all other transcripts of same species and shorter length
       If there is no overlap, remove the shorter fragment.

    Further filters include:
       * remove the shorter gene, if two genes overlap on the genome
         This is like a sanity check.
       
    """

    ###############################################################
    ###############################################################
    ###############################################################        
    ## identify all split transcripts
    ###############################################################
    removed_transcripts = set()

    if options.loglevel >= 2:
        options.stdlog.write("# removeSplitTranscripts started.\n" )
    
    for species, transcripts in map_species2transcripts.items():

        transcripts.sort()
        transcripts.reverse()

        for t1 in range(len(transcripts)-1):
            transcript1 = transcripts[t1][1]

            for t2 in range(t1+1, len(transcripts)):
                transcript2 = transcripts[t2][1]
                
                overlap12 = getPercentOverlap( mali[transcript1], mali[transcript2] )

                if options.loglevel >= 2:
                    options.stdlog.write("# %s %s: overlap=%5.2f\n" % (transcript1,transcript2,overlap12))
                
                if exons and hasGenomicOverlap( transcript1, transcript2, exons ):
                    if options.loglevel >= 1:
                        options.stdlog.write("# transcript: %s removed because genomic overlap with %s\n" % (transcript2, transcript1) )
                    removed_transcripts.add( (transcript2, transcript1, "genomic overlap") )
                    continue
                
                if exons:
                    
                    genomic_distance = getGenomicDistance( transcript1, transcript2, exons )
                    is_synchronous = isSynchronous( transcript1, transcript2, mali, exons )
                    
                    if genomic_distance > 0 and genomic_distance < options.min_genomic_distance and \
                           overlap12 < options.min_percent_overlap and \
                           is_synchronous:
                        if options.loglevel >= 1:
                            options.stdlog.write("# transcript: %s removed because genomic proximity to %s: %i/%5.2f\n" % (transcript2, transcript1, genomic_distance, overlap12) )
                        removed_transcripts.add( (transcript2, transcript1, "genomic proximity: %i/%5.2f%%" % (genomic_distance, overlap12) ) )
                        continue
                    
                if overlap12 < options.max_percent_overlap:
                    if options.loglevel >= 1:
                        options.stdlog.write("# transcript: %s removed because no overlap with %s\n" % (transcript2, transcript1) )
                    removed_transcripts.add( (transcript2, transcript1, "no mali overlap" ) )
                    continue
                    
    return removed_transcripts

##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------
##-------------------------------------------------------------------------------    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: mali2cleaned_mali.py 2781 2009-09-10 11:33:14Z andreas $")

    parser.add_option("-m", "--genome-master", dest="genome_master", type="string",
                      help="genome to use as master." )

    parser.add_option("-s", "--filename-removed", dest="filename_removed", type="string",
                      help="output filename for deleted entries." )

    parser.add_option("-e", "--filename-exons", dest="filename_exons", type="string",
                      help="filename on where to exon information."  )

    parser.add_option("-u", "--filename-summary", dest="filename_summary", type="string",
                      help="output filename of component summary." )

    parser.add_option("-c", "--filename-components", dest="filename_components", type="string",
                      help="output filename for components." )

    parser.add_option("--min-percent-overlap", dest="min_percent_overlap", type="float",
                      help="minimum percent overlap for splitting multiple alignment into components.")

    parser.add_option("--max-percent-overlap", dest="max_percent_overlap", type="float",
                      help="maximum percent overlap for split genes.")

    parser.add_option("--min-genomic-distance", dest="min_genomic_distance", type="int",
                      help="minimum genomic distance for adjacent genes to be considered dodgy.")

    parser.add_option("-o", "--mode", dest="mode", type="choice", 
                      choices=("joining", "split"),
                      help="""how to filter the alignment.
                      joining: remove joining transcripts (spindly genes)
                      split:  remove split transcripts""")

    parser.add_option("-g", "--gene-mode", dest="gene_mode", action="store_true",
                      help="""the aligned sequences are genes. This forces the exon boundaries to
                      collated by genes.""")
    
    parser.set_defaults( \
        genome_master = None,
        filename_removed = None,
        filename_components = None,        
        filename_summary = None,
        filename_exons = None,
        mode="joining",
        input_format = "fasta",
        output_format = "fasta", 
        max_percent_overlap = 0,
        min_percent_overlap = 0,
        gene_mode = False,
        separator = "|")
        
    (options, args) = E.Start( parser )

    ###############################################################
    ###############################################################
    ###############################################################        
    ## input
    ###############################################################                


    mali = Mali.Mali()
    mali.readFromFile( sys.stdin, format = options.input_format )
    all_identifiers = mali.getIdentifiers()

    if options.filename_exons:
        ## read exon boundaries and keep forward coordinates
        
        if options.gene_mode:
            exons = Exons.ReadExonBoundaries( open(options.filename_exons, "r"),
                                              from_zero = True)
            
            gene_exons = {}
            for id, ee in exons.items():
                data = id.split(options.separator)
                new_id = options.separator.join((data[0], data[2]))
                if new_id not in gene_exons: gene_exons[new_id] = []
                for e in ee:
                    e.mQueryToken = new_id
                gene_exons[new_id] += ee
            for id, ee in gene_exons.items():
                ee.sort( lambda x,y: cmp( x.mGenomeFrom, y.mGenomeFrom ) )
            exons = gene_exons

        else:
            exons = Exons.ReadExonBoundaries( open(options.filename_exons, "r"),
                                              filter=set(all_identifiers),
                                              from_zero = True)
            
    else:
        exons = {}

    ###############################################################
    ###############################################################
    ###############################################################        
    ## collect all transcripts for a species together with their
    ## aligned length
    ###############################################################                
    map_species2transcripts = {}
    
    for id in mali.getIdentifiers():
        data = id.split(options.separator)

        species = data[0]

        if exons:
            l = exons[id][-1].mGenomeTo - exons[id][0].mGenomeFrom
        else:
            l = len(mali.getEntry(id).getSequence())
        
        try:
            map_species2transcripts[species].append( (l, id ) )
        except KeyError:
            map_species2transcripts[species]= [( l, id) ]

    if options.mode == "joining":
        mapped_transcripts = removeJoiningTranscripts( mali, exons, map_species2transcripts, options )
        
    elif options.mode == "split":
        mapped_transcripts = removeSplitTranscripts( mali, exons, map_species2transcripts, options )

    ###############################################################
    ###############################################################
    ###############################################################        
    ## now build overlap graph of remaining sequences split multiple
    ## alignment in components.
    ## Compute reciprocal best match graph
    ###############################################################                
    graph = networkx.Graph()

    removed_transcripts = set(map(lambda x: x[0], mapped_transcripts ))

    for t in all_identifiers:
        if t not in removed_transcripts:
            graph.add_node( t )

    for t1 in range(len(all_identifiers)-1):
        transcript1 = all_identifiers[t1]
        if transcript1 in removed_transcripts: continue
        
        for t2 in range(t1+1, len(all_identifiers)):
            transcript2 = all_identifiers[t2]            
            if transcript2 in removed_transcripts: continue

            overlap = getPercentOverlap( mali[transcript1], mali[transcript2] )
            if overlap > 5:
                graph.add_edge( transcript1, transcript2 )
                
    ## compute components
    components = networkx.connected_components( graph )

    ###############################################################
    ###############################################################
    ###############################################################        
    ## output
    ###############################################################                
    if options.filename_components:
        n = 1
        outfile = open(options.filename_components, "w" )
        
        outfile.write( "id\tcomponent\n" )    
        for component in components:
            for c in component:
                outfile.write( "%s\t%i\n" % (c, n))
            n += 1
        outfile.close()
        
    if options.filename_removed and len(removed_transcripts) > 0:
        outfile = open(options.filename_removed, "w")
        outfile.write("removed\trepresentative\treason\n")
        for removed_transcript, rep_transcript, reason in mapped_transcripts:
            outfile.write("%s\t%s\t%s\n" % (removed_transcript, rep_transcript, reason) )
        outfile.close()

    if options.filename_summary:
        n = 1
        outfile = open(options.filename_summary, "w")
        outfile.write("component\tsize\tnspecies\tnmaster\n" )
        for component in components:
            species = map( lambda x: x.split(options.separator)[0], component)
            outfile.write( "%i\t%i\t%i\t%i\t%i\n" % ( n,
                                                      len(component),
                                                      len(species),
                                                      len(filter( lambda x: x == options.genome_master, species))))

            n += 1

    for transcript in removed_transcripts:
        mali.deleteEntry( transcript )

    new_identifiers = mali.getIdentifiers()
    
    mali.removeGaps( minimum_gaps = len( new_identifiers) )

    mali.writeToFile( options.stdout, format = options.output_format )
        
    if options.loglevel >= 1:
        options.stdlog.write( "# input=%i, output=%i, removed=%i, ncomponents=%i\n" % (len(all_identifiers), len(new_identifiers), len(removed_transcripts), len(components) ))
        options.stdlog.write( "# final component sizes: %s\n" % ",".join(map( lambda x: str(len(x)), components)))

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Predictor.py - running gene predictions with exonerate
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import os, sys, string, re, optparse, tempfile, time, subprocess

USAGE="""python %s [OPTIONS] peptide genome

Version: $Id: Predictor.py 2781 2009-09-10 11:33:14Z andreas $

Wrapper for running gene predictor on sequence chunks.

Format of gff file:

The last (9th) column containing the query identifier and an optional id. For example:

chr1    pseudogene.org  pseudogene (Processed)  31282205        31282680        .       +       .       Query Q9Y3D8; Id 74
chr1    pseudogene.org  pseudogene (Duplicated) 33225155        33225823        .       +       .       Query P15880; Id 77
chr1    pseudogene.org  pseudogene (Duplicated) 33226224        33226423        .       +       .       Query P15880; Id 77

or just

chr1    pseudogene.org  pseudogene (Processed)  31282205        31282680        .       +       .       Q9Y3D8
chr1    pseudogene.org  pseudogene (Duplicated) 33225155        33225823        .       +       .       P15880
chr1    pseudogene.org  pseudogene (Duplicated) 33226224        33226423        .       +       .       P15880

The file has to be sorted appropriately so that all the regions belonging to a group are adjacent.
"""

import Experiment as E
import Genomics
import IndexedFasta
import Prediction
import PredictionParser
try: import alignlib
except ImportError: pass
import Cluster
import GTF
import threading

class Error(Exception):
    """Base class for exceptions in this module."""
    def __str__(self):
        return str(self.message)
    def _get_message(self, message): return self._message
    def _set_message(self, message): self._message = message
    message = property(_get_message, _set_message)

class ParsingError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class Predictor:

    mLogLevel = 0
    
    def __init__(self):

        self.mDryRun = False

        self.mCluster = None

    def __call__( self,
                  filename_peptides,
                  filename_genome,
                  options = None):
        """run a single prediction using prediction program.

        return lines
        """

        options += " " + self.mInputOptions

        if not options:
            options += " " + self.mDefaultOptions

        statement = self.buildStatement( filename_peptides, filename_genome, options )

        if self.mLogLevel >= 4:
            print "# statement: %s" % statement
            sys.stdout.flush()

        if self.mDryRun:
            print statement
            return Prediction.Predictions()

        if self.mCluster:
            out, err = self.mCluster.runJob( statement )
        else:
            s = subprocess.Popen( statement,
                                  shell = True,
                                  stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE,
                                  close_fds = True)                              

            (out, err) = s.communicate()

            if s.returncode != 0:
                print err
                print out
                raise IOError, "Error in running statement %s" % statement
        
        if self.mLogLevel >= 3:
            print"# received %i chars from %s" % (len(out), self.mExecutable)
            print out            

        results = self.mParser.parse( map( lambda x: x+ "\n", out.split("\n") ) )
        return results

    def setCluster( self, cluster ):
        """set cluster object."""
        self.mCluster = cluster

    def setOutputOptions(self, options):
        """set output options."""
        self.mOutputOptions = options 

    def setDryRun( self, flag = True):
        """setup dry run.

        Gene prediction is not run, but files are set up as if it were.
        """
        self.mDryRun = flag

    def setLogLevel( self, level ):
        """set loglevel."""
        self.mLogLevel = level


    def getParser( self ):
        """returns the parser object."""
        return self.mParser

    def predictFromSequences( self,
                              peptide_sequence,
                              genomic_sequence,
                              options = None ):
        """predict from two explicitely given sequences."""

        outfile, filename_peptides = tempfile.mkstemp()
        os.write( outfile, ">query\n%s" % peptide_sequence )
        os.close(outfile)

        outfile, filename_genome = tempfile.mkstemp()
        os.write( outfile, ">target\n%s" % genomic_sequence )
        os.close(outfile)

        self.mParser.addGenomicSequence( "target", genomic_sequence )
        result = self.__call__( filename_peptides, filename_genome, options )
        self.mParser.deleteGenomicSequence( "target" )
        os.remove( filename_peptides )
        os.remove( filename_genome )
        
        return result

    def predictFromRange( self,
                          peptide_sequence,
                          contig, strand, start, end,
                          fasta,
                          options = None ):
        """predict from a peptide sequence within a region."""

        
        genomic_sequence = fasta.getSequence( contig, strand, start, end )

        result = self.predictFromSequences( peptide_sequence, genomic_sequence, options )

        lcontig = fasta.getLength( contig )

        for r in result:

            # shift aligned region for fragment
            # this is strand dependent.
            r.shiftGenomicRegion( start, lcontig - start )
            ## set sbjct token to original
            r.mSbjctToken = contig
            
            ## sort out strand
            ## if match is on negative strand
            if Genomics.IsNegativeStrand( r.mSbjctStrand ):
                ## switch coordinates
                if Genomics.IsNegativeStrand( strand ):
                    strand = "+"
                else:
                    strand = "-"
                    
            r.mSbjctStrand = strand

        return result

class PredictorGenewise(Predictor):
    """run genewise as predictor.

    genewise only runs a single peptide sequence against a single
    DNA sequences. genewisedb runs collections of peptide sequences
    against collections of DNA sequences. Note that this might take a while
    so it is prudent to spread it over the cluster.
    """
    
    def __init__( self, parser = None ):
        Predictor.__init__(self)
        if parser:
            self.mParser = parser
        else:
            self.mParser = PredictionParser.PredictionParserGenewise()
            
        self.mExecutable = "genewise"
        self.mDefaultOptions="-pseudo -init endbias"
        self.mOutputOptions="-quiet -sum -gff -trans -pep -alb "
        self.mInputOptions = ""

    def buildStatement( self, filename_peptides, filename_genome, options ):
        return string.join( map(str, (
            self.mExecutable,
            options,
            self.mOutputOptions,
            filename_peptides,
            filename_genome,
            )), " " )

class PredictorGenewiseHMM(Predictor):
    
    def __init__( self, parser = None ):
        Predictor.__init__(self)
        if parser:
            self.mParser = parser
        else:
            self.mParser = PredictionParser.PredictionParserGenewiseHMM()
            
        self.mExecutable = "genewise -hmmer"
        self.mDefaultOptions="-pseudo -alg 623L "
        self.mInputOptions = ""
        self.mOutputOptions="-quiet -sum -gff -trans -pep -alb "

    def buildStatement( self, filename_peptides, filename_genome, options ):

        ## allow parser to extract query length from the
        ## submitted hmm.
        self.mParser.extractQueryLength( filename_peptides )
        
        return string.join( map(str, (
            self.mExecutable,
            options,
            self.mOutputOptions,
            filename_peptides,
            filename_genome,
            )), " " )
                        
class PredictorExonerate ( Predictor ):

    def __init__( self, parser = None ):
        
        Predictor.__init__(self)
        
        self.mExecutable = "exonerate"

        self.mInputOptions = "-Q protein -T dna --softmasktarget TRUE"
        self.mDefaultOptions = "-m p2g --forcegtag TRUE"
        self.mOutputOptions = '--forwardcoordinates FALSE --showalignment FALSE --showvulgar FALSE --showsugar FALSE --showcigar FALSE ' + \
                              ' --ryo "diy\t%S\t%ql\t%r\t%pi\t%ps\t%V\n" --showtargetgff FALSE --showquerygff FALSE'
        
        ## only predictions on this strand
        self.mRestrictToStrand = None

        if parser:
            self.mParser = parser
        else:
            self.mParser = PredictionParser.PredictionParserExonerate()

    def buildStatement( self, filename_peptides, filename_genome, options ):
        return string.join( map(str, (
            self.mExecutable,
            options,
            self.mOutputOptions,
            "--query",
            filename_peptides,
            "--target",
            filename_genome,
            )), " " )

##--------------------------------------------------------------------------
class PredictionParser :
    """base class of prediction parsers."""
    mLogLevel = 0
    
    def __init__( self,
                  peptides = None,
                  peptide = None,
                  genomes = None,
                  genome = None,
                  filename_genome = None,
                  filename_peptides = None,
                  indexed_genome = None,
                  ):
        """pass sequence data."""
        self.mPeptides = peptides
        self.mPeptide = peptide
        self.mGenomes = genomes
        self.mGenome = genome        
        self.mIndexedGenome = indexed_genome
        self.mFilenamePeptides = filename_peptides
        self.mFilenameGenome = filename_genome

        ## stop codons to ignore at boundaries
        self.mBorderStopCodons = 0

    def parse( self, lines ):
        """parse lines.
        """
        return Prediction.Predictions()

    def getPeptideSequence( self, query_token ):
        """returns peptide sequence for query_token."""


        if self.mPeptides:
            return self.mPeptides[query_token]
        elif self.mFilenamePeptides:
            self.mPeptides = Genomics.ReadPeptideSequences( open( self.mFilenamePeptides, "r") )
            return self.mPeptides[query_token]
        elif self.mPeptide:
            return self.mPeptide

    def addGenomicSequence( self, sbjct_token, sequence, strand = "+" ):
        """add a sequence so that the parser nkows about it.
        """
        if strand == "-":
            raise "not implemented, please add positive strand"

        if self.mGenomes:
            if sbjct_token in self.mGenomes:
                raise KeyError, "key %s already in genomes" % (sbjct_token)
            self.mGenomes[sbjct_token] = sequence
        elif self.mFilenameGenome:
            raise "not implemented"
        elif self.mIndexedGenome:
            raise "not implemented"
        else:
            self.mGenome = sequence

    def deleteGenomicSequence( self, sbjct_token ):
        """remove a genomic sequence from parser."""
        
        if self.mGenomes:
            if sbjct_token in self.mGenomes:
                del self.mGenomes[sbjct_token]
        elif self.mFilenameGenome:
            raise "not implemented"
        elif self.mIndexedGenome:
            raise "not implemented"

    def getGenomicSequence( self, sbjct_token, sbjct_strand = None, sbjct_from = None, sbjct_to = None ):
        """returns a genomic sequence for given coordinates."""

        if self.mGenomes:
            genomic_sequence = self.mGenomes[sbjct_token]
        elif self.mFilenameGenome:
            self.mGenomes = Genomics.ReadGenomicSequences( open(self.mFilenameGenome, "r"), do_reverse = False )
            genomic_sequence = self.mGenomes[sbjct_token]
        elif self.mIndexedGenome:
            return self.mIndexedGenome.getSequence( sbjct_token, sbjct_strand, sbjct_from, sbjct_to )
        else:
            genomic_sequence = self.mGenome

        if sbjct_strand in ("-", "0", "-1", -1):
            genomic_sequence = string.translate( genomic_sequence, string.maketrans("ACGTacgt", "TGCAtgca") )[::-1]

        if sbjct_from != None and sbjct_to != None:
            return genomic_sequence[sbjct_from:sbjct_to]
        else:
            return genomic_sequence

    def getGenomicSequenceLength( self, sbjct_token ):
        """returns a genomic sequence for given coordinates."""

        if self.mGenomes:
            return len(self.mGenomes[sbjct_token])
        elif self.mFilenameGenome:
            self.mGenomes = Genomics.ReadGenomicSequences( open(self.mFilenameGenome, "r") )
            return len(self.mGenomes[sbjct_token])            
        elif self.mIndexedGenome:
            return self.mIndexedGenome.getLength( sbjct_token )
        else:
            return len(self.mGenome)
    
##--------------------------------------------------------------------------
class PredictionParserGenewise( PredictionParser ):
    """Parse output from genewise.
    """
    
    def __init__(self, **kwargs):
        
        PredictionParser.__init__( self, **kwargs )

        self.mParsers = [ self.parseSummary, self.parseTranslation,  self.parsePeptide, self.parseGFF, self.parseAlb ]

        self.mQueryLength = None
        
    def parse( self, lines ):
        """call the various parsers for a chunk of data.
        """
        if len(lines) == 0: return 0

        matches = Prediction.Predictions()
        self.mMatches = matches
        self.mTranslations = []
        self.mParsedPeptides = []
        self.mAlignments = []
        self.mSummaries = []
        self.strands = []

        # for line in lines:
        # print line,
        
        ###############################################################################
        ## call the various parsers for chunks separated by "//"
        chunks = filter( lambda x: lines[x][:2] == "//", range( len(lines )) )

        if len(chunks) % len(self.mParsers) != 0:

            for x in range(len(chunks)-1):
                print "#########################################################"
                print "".join(lines[chunks[x]+1:chunks[x+1]])
            
            raise ParsingError( "lengths of sections do not not fit: found %i, expected %i" % (len(chunks), len(self.mParsers)))

        chunks = [-1] + chunks
        
        chunk = 0
        while chunk < len(chunks) - 1:
            
            for parser in self.mParsers:
                parser(lines[chunks[chunk]+1:chunks[chunk+1]])
                chunk += 1

        ###############################################################################
        # build the Prediction entries
        for x in range(len(self.mSummaries)):
            
            entry = Prediction.Prediction()

            summary = self.mSummaries[x]

            (entry.mQueryToken, entry.mSbjctToken, entry.score,
             entry.mQueryFrom, entry.mQueryTo, 
             entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
             entry.mNIntrons) = \
                            (summary.mQueryToken, summary.mSbjctToken, summary.score,
                             summary.mQueryFrom, summary.mQueryTo, 
                             summary.mSbjctGenomeFrom, summary.mSbjctGenomeTo,
                             summary.mIntrons)

            entry.mMapPeptide2Genome, entry.mMapPeptide2Translation, nmatches, nindels = self.mAlignments[x]
            
            ## in case of pseudogenes, translations are empty, but peptides are there,
            ## so using those.
            entry.mTranslation = self.mParsedPeptides[x][1]

            entry.mSbjctStrand = self.strands[x]
            peptide_sequence = self.getPeptideSequence( entry.mQueryToken )

            if peptide_sequence:
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                alignlib.rescoreAlignment( entry.mMapPeptide2Translation, row_seq, col_seq )

                entry.mQueryLength = len(peptide_sequence)            
                entry.mPercentIdentity = alignlib.calculatePercentIdentity( entry.mMapPeptide2Translation, row_seq, col_seq ) * 100
                entry.mPercentSimilarity = alignlib.calculatePercentSimilarity( entry.mMapPeptide2Translation ) * 100
                entry.mQueryCoverage = ( entry.mMapPeptide2Translation.getRowTo() - \
                                         entry.mMapPeptide2Translation.getRowFrom() + 1 ) * 100 /\
                                         entry.mQueryLength
            elif self.mQueryLength:
                ## for hmms, query lenght can be extraced from the HMM
                entry.mQueryLength = self.mQueryLength
                entry.mQueryCoverage = ( entry.mMapPeptide2Translation.getRowTo() - \
                                         entry.mMapPeptide2Translation.getRowFrom() + 1 ) * 100 /\
                                         entry.mQueryLength
                ## percent identity is given by the number of matches versus alignment length
                entry.mPercentIdentity = float(nmatches) / (nmatches + nindels)
                entry.mPercentSimilarity = entry.mPercentIdentity

            entry.mAlignmentString = string.join( map( \
                                      lambda x: string.join(map(str, x), " "),
                                      entry.mMapPeptide2Genome), " ")

            if len(entry.mMapPeptide2Genome) == 0:
                print "### PARSING ERROR: empty alignment"
                print str(entry)
                print string.join(lines, "")
                sys.exit(1)

            # get regions matching
            # use summary information, as it is correct for genomic data.
            # when -u is set (in contrast to -alb format)
            # genewise starts counting at 1, thus subtract 1 from first position
            # Note: When position starts with a gap, genewise says alignments starts
            # from position 1 while the first aligned residue is position 2
            if entry.mMapPeptide2Genome[0][0] == "G": entry.mQueryFrom += entry.mMapPeptide2Genome[0][1]
            if entry.mMapPeptide2Genome[-1][0] == "G": entry.mQueryTo -= entry.mMapPeptide2Genome[0][1]

            # a sanity check
            if entry.mQueryFrom != entry.mMapPeptide2Translation.getRowFrom():
                print "## PARSING ERROR: wrong start at %s vs %s: %i <> %i" %\
                      (entry.mQueryToken, entry.mSbjctToken,
                       entry.mQueryFrom,
                       entry.mMapPeptide2Translation.getRowFrom() )
                print str(entry)
                print string.join(lines,"")
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                print str( alignlib.AlignmentFormatExplicit( entry.mMapPeptide2Translation, row_seq, col_seq ))
                sys.exit(1)
            if entry.mQueryTo != entry.mMapPeptide2Translation.getRowTo() and \
                   entry.mMapPeptide2Genome[-1][0] != "G":
                print "## PARSING ERROR: wrong end at %s vs %s: %i <> %i" %\
                  (entry.mQueryToken, entry.mSbjctToken,
                   entry.mQueryTo,
                   entry.mMapPeptide2Translation.getRowTo() )
                print str(entry)                
                print string.join(lines,"")
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                print str( alignlib.AlignmentFormatExplicit( entry.mMapPeptide2Translation, row_seq, col_seq ))
                sys.exit(1)

            ## fix coordinates
            
            ## sic: on neg strand genewise has high,low coordinates
            if entry.mSbjctStrand == "-":
                lgenome = self.getGenomicSequenceLength( entry.mSbjctToken )                
                entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo = lgenome - entry.mSbjctGenomeFrom, lgenome - entry.mSbjctGenomeTo
                entry.mSbjctGenomeTo += 1
            else:
                entry.mSbjctGenomeFrom -= 1

            genomic_sequence = self.getGenomicSequence( entry.mSbjctToken, entry.mSbjctStrand,
                                                        entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo)

            if genomic_sequence:
                (entry.mNIntrons, entry.mNFrameShifts, entry.mNGaps, entry.mNSplits, entry.mNStopCodons, disruptions) = \
                                  Genomics.CountGeneFeatures( 0,
                                                              entry.mMapPeptide2Genome,
                                                              genomic_sequence,                                                          
                                                              self.mBorderStopCodons )
                
            
            matches.append( entry )
            

        matches.sort( lambda x,y : cmp( -x.score, -y.score))
        
        return matches
    
    def parseGFF( self, lines ):
        """extract strands from the gff section."""
        for line in lines:
            data = re.split( "\s+", line[:-1])
            if data[2] == "match":
                self.strands.append( data[6] )

    def parseSummary( self, lines ):
        """parse summary entry of genewise."""

        class SummaryResult:
            pass

        ## first line is header
        for line in lines[1:]:
            data = re.split( "\s+", line[:-1])
            entry = SummaryResult()
            entry.score = float(data[0])
            entry.mQueryToken = data[1]
            entry.mQueryFrom = int(data[2])
            entry.mQueryTo = int(data[3])
            entry.mSbjctToken = data[4]
            entry.mSbjctGenomeFrom = int(data[5])
            entry.mSbjctGenomeTo = int(data[6])
            entry.mIndels = int(data[7])
            if data[8] == "":
                entry.mIntrons = 0
            else:
                entry.mIntrons = int(data[8])
                
            self.mSummaries.append( entry )

    def parseTranslation( self, lines ):
        """read a translation entry from genewise.        
        """
        ## There might be some problems with multiple entries.
        ##
        self.mTranslations += self.__parseFasta( lines )

    def parsePeptide( self, lines ):
        """read a peptide entry from genewise.
        """
        self.mParsedPeptides += self.__parseFasta( lines )
        
    def __parseFasta( self, lines ):
        """read a fasta entry from genewise.

        stores tuples of description,sequence in result.
        """

        results = []
        fragments = []
        description = ""
        for line in lines:
            if line[0] in (">", "#"):
                if fragments:
                    results.append( (description, "".join(fragments) ) )
                    
                if line[0] != "#":
                    description = line[1:-1]
                else:
                    description = "pseudogene"
                    
                fragments = []
                continue
            elif line[:6] == "Making":
                continue
            fragments.append(line[:-1])

        if fragments:
            results.append( ( description, "".join(fragments) ) )

        return results
        
    def parseAlb( self, lines ):
        """Parse genewise logical block alignment and return aligment.

        ( alignment of peptide to genome,
          alignment of peptide to translation )

        """

        map_genewise2code = \
                          { ( "MATCH_STATE", "CODON") : "M",
                            ( "INTRON_STATE", "5SS_PHASE_0"): "5",
                            ( "INTRON_STATE", "5SS_PHASE_1"): "5",                        
                            ( "INTRON_STATE", "5SS_PHASE_2"): "5",
                            ( "INTRON_STATE", "CENTRAL_INTRON"): "I",
                            ( "MATCH_STATE", "3SS_PHASE_0"): "3",
                            ( "MATCH_STATE", "3SS_PHASE_1"): "3",
                            ( "MATCH_STATE", "3SS_PHASE_2"): "3",                        
                            ( "INSERT_STATE", "3SS_PHASE_0") : "3",
                            ( "INSERT_STATE", "3SS_PHASE_1") : "3",
                            ( "INSERT_STATE", "3SS_PHASE_2") : "3",
                            ( "INTRON_MATCH_1", "5SS_PHASE_0"): "5",
                            ( "INTRON_MATCH_1", "5SS_PHASE_1"): "5",
                            ( "INTRON_MATCH_1", "5SS_PHASE_2"): "5",                            
                            ( "INTRON_MATCH_1", "CENTRAL_INTRON"): "I",
                            ( "INTRON_MATCH_1", "PYRIMIDINE_TRACT"): "I",
                            ( "DELETE_STATE", "INSERT") : "G",
                            ( "INSERT_STATE", "CODON") : "G",
                            ( "INSERT_STATE", "SEQUENCE_DELETION") : "F", # a codon with less than 3 nucleotides at an insert
                            ( "MATCH_STATE", "SEQUENCE_INSERTION") : "F", # a codon with more than 3 nucleotides
                            ( "MATCH_STATE", "SEQUENCE_DELETION")  : "F", # a codon with less than 3 nucleotides
                            ( "LOOP_STATE", "RANDOM_SEQUENCE") : None,
                            ( "GENOMIC_RND_STATE", "RANDOM_SEQUENCE") : None,
                            ( "GENOMIC_RND_STATE", "CODON" ): None,
                            ( "END", "END") : None,
                            }


        ## note: all the numbers including the ranges can be negative (-1 for first position) !
        rx = re.compile( "^(\S+)\s+\[(\S+):(\S+)\s+\"(\S+)\"\s+(\S+)\],\[(\S+):(\S+)\s+\"(\S+)\"\s+(\S+)\]" )
        
        query_last_state = None
        query_first_from = 0

        sbjct_last_state = None
        sbjct_genomic_residue_first_from = 0
        sbjct_peptide_residue_first_from = 0

        states = []
        total_bin_score = 0.0

        nintrons = 0
        nframeshifts = 0
        nphases = [0,0,0]
        ngaps = 0

        ## parsing errors
        nerrors = 0

        for line in lines:

            try:
                (bin_score,
                 query_from, query_to, query_state, query_score,
                 sbjct_from, sbjct_to, sbjct_state, sbjct_score ) =  rx.match( line[:-1] ).groups()
            except AttributeError:
                raise ParsingError, line

            query_from, query_to, sbjct_from, sbjct_to = map(int, (query_from, query_to, sbjct_from, sbjct_to))

            total_bin_score += string.atof( bin_score )

            if query_last_state != query_state or sbjct_last_state != sbjct_state:

                if query_last_state:
                    states.append( ( query_last_state, query_first_from, query_from,
                                     sbjct_last_state, sbjct_first_from, sbjct_from ) )

                
                query_first_from = query_from
                sbjct_first_from = sbjct_from
                query_last_state = query_state
                sbjct_last_state = sbjct_state

        states.append( ( query_last_state, query_first_from, query_to,
                         sbjct_last_state, sbjct_first_from, sbjct_to) )

        ## translate genewise alignment to exonerate alignment codes
        n = 0
        
        ## residue positions on the query/sbjct for peptide sequences
        ## note: genewise would start the first state at -1:0. Thus, add
        ## 2 to get first aligned position in peptide
        query_peptide_from = states[0][1] + 2
        sbjct_peptide_from = 1

        map_peptide2genome = []
        map_query2sbjct = alignlib.makeAlignmentVector()

        ## count number of matches/indels as proxy for pid
        ## See above for a loop, where counting could be done
        ## by scores.
        nmatches, nindels = 0, 0
        
        for x in range( len(states)):

            query_state, query_from, query_to, sbjct_state, sbjct_from, sbjct_to = states[x]

            query_len = query_to - query_from
            sbjct_len = sbjct_to - sbjct_from
            
            ## build alignment between peptide sequence and genomic sequence
            try:
                code = map_genewise2code[ (query_state, sbjct_state)]
            except KeyError:
                print "## ERROR: unknown genewise code: %s %s" % (query_state, sbjct_state)
                nerrors += 1
                code = "?"

            if query_state == "LOOP_STATE" and x > 0:
                if map_peptide2genome:
                    self.mAlignments.append( (map_peptide2genome, map_query2sbjct, nmatches, nindels) )
                map_query2sbjct = alignlib.makeAlignmentVector()
                map_peptide2genome = []
                query_peptide_from = states[x+1][1] + 2
                sbjct_peptide_from = 1
                nmatches = 0
                nindels = 0
                continue

            ## add code for split codons
            ## move matches to second split codon, if
            ## a split codon is on a gap: add a Gap
            if sbjct_state == "5SS_PHASE_1":
                map_peptide2genome.append( ("S", 0, 1) )
                sbjct_len -= 1
            elif sbjct_state == "5SS_PHASE_2":
                map_peptide2genome.append( ("S", 0, 2) )
                sbjct_len -= 2
            elif sbjct_state == "3SS_PHASE_1":
                query_len = 0
                sbjct_len -= 2
            elif sbjct_state == "3SS_PHASE_2":
                query_len = 0                
                sbjct_len -= 1

            if code:
                if query_from != query_peptide_from - 2:
                    print "Mismatch", query_state, query_from, query_to, sbjct_state,
                    sbjct_from, sbjct_to, query_peptide_from
                
                map_peptide2genome.append( (code, query_len, sbjct_len) )

            if sbjct_state == "3SS_PHASE_1":
                map_peptide2genome.append( ("S", 1, 2) )
                sbjct_len = 2
                if query_state == "MATCH_STATE":
                    query_len = 1
                else:
                    query_len = 0
                    
            elif sbjct_state == "3SS_PHASE_2":
                map_peptide2genome.append( ("S", 1, 1) )                
                sbjct_len = 1
                if query_state == "MATCH_STATE":
                    query_len = 1
                else:
                    query_len = 0
                
            ## build alignment between peptide sequences
            query_increment = 0
            sbjct_increment = 0

            if query_state == "MATCH_STATE":
                 query_increment = query_len
                 nmatches += query_len
                 
            elif query_state == "DELETE_STATE":
                query_increment = query_len
                nindels += query_len
                
            elif query_state == "INSERT_STATE":
                query_increment = query_len
                nindels += query_len
                
            elif query_state == "UNKNOWN_LABEL":
                query_peptide_from += query_len
                ## there are three X's in the sbjct
                sbjct_peptide_from += 3             
                continue
            elif query_state == "LOOP_STATE":
                query_increment += query_len

            if sbjct_state == "CODON":
                sbjct_increment = sbjct_len 
            elif sbjct_state in ("3SS_PHASE_1", "3SS_PHASE_2"):
                sbjct_increment = 3
            elif sbjct_state == "INSERT":
                sbjct_increment = sbjct_len
            elif sbjct_state == "RANDOM_SEQUENCE":
                sbjct_increment = 0

            ## for a frameshift: genewise outputs two characters in the
            ## translated sequence? I do not know. 3 makes more sense.
            elif sbjct_state == "SEQUENCE_DELETION":
                sbjct_increment = 3

            ## for a frameshift: if there is a sequence insertion, no
            ## character is output.
            elif sbjct_state == "SEQUENCE_INSERTION":
                if sbjct_len > 3:
                    sbjct_increment = 3
                else:
                    sbjct_increment = 0

            if code and query_increment and sbjct_increment:
                alignlib.addDiagonal2Alignment( map_query2sbjct, 
                                               query_peptide_from, query_peptide_from + query_increment - 1,
                                               sbjct_peptide_from - query_peptide_from
                                               )
                
            query_peptide_from += query_increment
            sbjct_peptide_from += sbjct_increment / 3
            
        if map_peptide2genome:
            self.mAlignments.append( (map_peptide2genome, map_query2sbjct, nmatches, nindels) )
    
##--------------------------------------------------------------------------
class PredictionParserGenewiseHMM( PredictionParserGenewise ):
    """Parse output from genewise.

    This subclass reads the profile from the file to extract
    the length of a query.
    """
    
    def __init__(self, **kwargs):
        
        PredictionParserGenewise.__init__( self, **kwargs )
        
    def extractQueryLength( self, filename_peptides ):
        """extract query length from HMM."""

        infile = open(filename_peptides, "r")
        for line in infile:
            if line[:4] == "LENG":
                self.mQueryLength = int(line[5:-1])
                break
        else:
            self.mQueryLength = None
        infile.close()
        
##--------------------------------------------------------------------------
class PredictionParserExonerate (PredictionParser):
    """Parse output from exonerate.
    """
    
    def __init__(self, **kwargs):
        PredictionParser.__init__( self, **kwargs )

        self.mRestrictToStrand = None

    def SetRestrictToStrand( self, v ):
        self.mRestrictToStrand = v
            
    def parse( self, lines ):

        matches = Prediction.Predictions()
        
        nmatches = 0
        for line in lines:
            
            if not re.match("diy", line): continue

            nmatches += 1
            data = string.split( line[:-1], "\t")

            entry = Prediction.Prediction()
            entry.mRank = nmatches 

            (sugar, entry.mQueryLength, rank, percent_identity, percent_similarity, vulgar ) = data[1:]

            (entry.mQueryToken, entry.mQueryFrom, entry.mQueryTo, entry.mQueryStrand, 
             entry.mSbjctToken, entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
             entry.mSbjctStrand, entry.score) = string.split(sugar, " " )

            if self.mRestrictToStrand and entry.mSbjctStrand != self.mRestrictToStrand:
                continue

            try:
                (entry.mQueryFrom, entry.mQueryTo,
                 entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                 entry.score, entry.mQueryLength) = map( \
                    int , (entry.mQueryFrom, entry.mQueryTo,
                           entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                           entry.score, entry.mQueryLength))
                
            except ValueError:
                raise InputError, "error while parsing ints: %s" % " ".join( (entry.mQueryFrom, entry.mQueryTo,
                                                                            entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                                                                            entry.score, entry.mQueryLength) )
            
            try:            
                entry.mPercentIdentity, entry.mPercentSimilarity = map( float,
                                                                        (percent_identity, percent_similarity ))
            except ValueError:
                raise InputError, "error while parsing floats: %s, %s" % (percent_identity, percent_similarity)
            
            entry.mAlignmentString = vulgar
            try:
                entry.mMapPeptide2Genome = Genomics.String2Alignment( vulgar )
            except ValueError:
                raise InputError, "error while parsing alignment %s" % vulgar
            
            entry.mQueryCoverage = (entry.mQueryTo - entry.mQueryFrom) * 100 / entry.mQueryLength

            genomic_sequence = self.getGenomicSequence( entry.mSbjctToken, entry.mSbjctStrand,
                                                        entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo )

            entry.mMapPeptide2Translation, entry.mTranslation = Genomics.Alignment2PeptideAlignment( \
                entry.mMapPeptide2Genome, entry.mQueryFrom, 0, genomic_sequence )

            ## add 1 for peptide alignments
            entry.mQueryFrom += 1

            if genomic_sequence:
                (entry.mNIntrons, entry.mNFrameShifts, entry.mNGaps, entry.mNSplits, entry.mNStopCodons, disruptions) = \
                                  Genomics.CountGeneFeatures( 0,
                                                              entry.mMapPeptide2Genome,
                                                              genomic_sequence,
                                                              self.mBorderStopCodons )                                                              

            matches.append(entry)
            
        return matches

##--------------------------------------------------------------------------
class PredictionParserEcho (PredictionParser):
    """simply output predictions - do not parse."""

    
    def __init__(self, **kwargs):
        PredictionParser.__init__( self, **kwargs )

    def parse( self, lines ):
        """call the various parsers for a chunk of data.
        """

        for line in lines:
            print line[:-1]

        return Prediction.Predictions()        

##--------------------------------------------------------------------------
class TandemPredictor:
    """run a tandem prediction

    If a match is found, look for further related matches upstream or
    downstream.

    This is a hack that works for one additional query downstream or
    upstream. To be fully workable, sort out reverse strand predictions
    and overlapping predictions. Also, make sure that if several results
    are obtained, they are not spanning each other.
    """

    def __init__(self, predictor,
                 upstream_queries = [],
                 downstream_queries = []):
        """run predictor on queries."""
        self.mPredictor = predictor

        self.mDownStreamQueries = downstream_queries
        self.mUpStreamQueries = upstream_queries        

        ## maximal search region for additional elements
        self.mMaxExtension = 5000
        ## minimum distance between consecutive elements
        self.mMinDistance  = 0

        ## minimum length region to search
        self.mMinGenomeSize = 100

        self.mLogLevel = 0

    def setLogLevel( self, level ):
        self.mLogLevel = level

    def runTandem( self,
                   filename_peptides,
                   filename_genome,
                   options,
                   sbjct_token, sbjct_strand,
                   start, end ):

        """run a tandem prediction with region given by start, end."""

        parser = self.mPredictor.getParser()
        lgenome = parser.getGenomicSequenceLength( sbjct_token )

        start = max( 0, start )
        end = min( end, lgenome )

        if end - start < self.mMinGenomeSize:
            return None

        if self.mLogLevel >= 2:
            sys.stdout.write("# running tandem prediction on fragment %s:%s:%i:%i\n" % (sbjct_token,
                                                                                        sbjct_strand,
                                                                                        start,
                                                                                        end))

        ## patch - tempdir is loosing temporary directory assignment
        ## why is that?
        tempfile.tempdir = "."
        new_outfile_genome, new_filename_genome = tempfile.mkstemp()

        ## Note: genewise fails with long contig names, thus
        ## truncate the path.
        key = os.path.basename(new_filename_genome)
        genomic_sequence =  parser.getGenomicSequence( sbjct_token,
                                                       sbjct_strand,
                                                       start, end )
        os.write(new_outfile_genome, ">%s\n%s\n" % (key, genomic_sequence) )

        
        os.close( new_outfile_genome )
        parser.addGenomicSequence( key, genomic_sequence )
                                   
        result = self.mPredictor( filename_peptides, new_filename_genome, options )
        result.shiftGenomicRegion( start, lgenome - end )
        
        for r in result:
            r.mSbjctToken = sbjct_token
            
        parser.deleteGenomicSequence( key )

        if self.mLogLevel < 2:
            os.remove( new_filename_genome )
        
        if len(result) == 0:
            return None
        else:
            # return only the best match
            return result.getBestMatch()
    
    def __call__( self,
                  filename_peptides,
                  filename_genome,
                  options = None):
        """call runRecursive."""

        ## do predictions to get anchors
        results = self.mPredictor( filename_peptides, filename_genome, options )

        if not results: return results

        ## get parser object
        parser = self.mPredictor.getParser()

        ## place to store new results
        new_results = Prediction.Predictions()

        ## sort results by location
        results.sort( lambda x,y: cmp( (x.mSbjctToken,x.mSbjctStrand,x.mSbjctGenomeFrom),
                                       (y.mSbjctToken,y.mSbjctStrand,y.mSbjctGenomeFrom) ) )


        last_sbjct_token, last_sbjct_strand = None, None
        
        for result in results:

            if self.mLogLevel >= 2:
                sys.stdout.write("# extending prediction: %s\n" % (str(result) ))

            sbjct_token, sbjct_strand = result.mSbjctToken, result.mSbjctStrand
            
            if last_sbjct_token != sbjct_token or last_sbjct_strand != sbjct_strand:
                
                lgenome = parser.getGenomicSequenceLength( sbjct_token )
                start = 0
                end   = lgenome

            last_sbjct_token = sbjct_token
            last_sbjct_strand = sbjct_strand
            
            upstream_results = Prediction.Predictions()

            start = max( start, result.mSbjctGenomeFrom - self.mMinDistance + self.mMaxExtension )

            for filename_query in self.mUpStreamQueries:
                r = self.runTandem( filename_query,
                                    filename_genome,
                                    options,
                                    sbjct_token, sbjct_strand,
                                    start,
                                    max( start, result.mSbjctGenomeFrom - self.mMinDistance) )

                if r:
                    if self.mLogLevel >= 2:
                        sys.stdout.write("# upstream extension successful: %s\n" % (str(r) ))
                                             
                    upstream_results.add( r )
                else:
                    if self.mLogLevel >= 2:
                        sys.stdout.write("# upstream extension unsuccessful.\n" )
                                             
            downstream_results = Prediction.Predictions()
            for filename_query in self.mDownStreamQueries:
                r = self.runTandem( filename_query,
                                    filename_genome,
                                    options,
                                    sbjct_token, sbjct_strand,
                                    min( lgenome, result.mSbjctGenomeTo + self.mMinDistance),
                                    min( lgenome, result.mSbjctGenomeTo + self.mMinDistance + self.mMaxExtension ) )

                if r:
                    if self.mLogLevel >= 2:
                        sys.stdout.write("# downstream extension successful: %s\n" % (str(r) ))
                                             
                    downstream_results.add( r )
                else:
                    if self.mLogLevel >= 2:
                        sys.stdout.write("# downstream extension unsuccessful.\n" )
                                             
            ## now sort out the predictions by combining anchors
            ## with upstream and downstream predictions
            if upstream_results:
                upstream_results.sort( lambda x,y: cmp( (x.mSbjctToken, x.mSbjctStrand, x.mSbjctGenomeFrom), (y.mSbjctToken, y.mSbjctStrand, y.mSbjctGenomeTo) ) )
                new_result = upstream_results[0]

                for r in upstream_results[1:]:
                    new_result.Add( r, combine_queries = True )
                    
                new_result.Add( result, combine_queries = True )
            else:
                new_result = result

            if downstream_results:
                downstream_results.sort( lambda x,y: cmp( (x.mSbjctToken, x.mSbjctStrand, x.mSbjctGenomeFrom), (y.mSbjctToken, y.mSbjctStrand, y.mSbjctGenomeTo) ) )

                for r in downstream_results:
                    new_result.Add( r, combine_queries = True )

            new_results.add( new_result )

            start = new_result.mSbjctGenomeTo

        return new_results
    
    
##--------------------------------------------------------------------------
class RecursivePredictor:
    """run recursive prediction.

    Note: this will always run the full peptide set against the matched fragments,
    so only use this if you have only a single input sequence/hmm and the matches
    returned are not repetetive.
    """
    
    def __init__(self, predictor ):

        self.mPredictor = predictor
        
        # analyze at least 500 kb
        self.mMinGenomeSize = 500

        self.mLogLevel = 0

        self.mLevel = 0

    def setLogLevel( self, level ):
        self.mLogLevel = level
        
    def runRecursive( self,
                      filename_peptides,
                      filename_genome,
                      options ):
        
        ## do predictions
        results = self.mPredictor( filename_peptides, filename_genome, options )

        self.mLevel += 1

        infile = open(filename_genome, "r" )
        genome = Genomics.ReadGenomicSequences( infile, do_reverse = False )
        infile.close()
        
        ## collect new fragments to run
        fragments = {}

        for result in results:
            lgenome = len(genome[result.mSbjctToken])
            if result.mSbjctGenomeFrom >= self.mMinGenomeSize:
                key = "%s:%s:%i:%i" % (result.mSbjctToken, result.mSbjctStrand, 0, result.mSbjctGenomeFrom)
                fragments[key] = ( result.mSbjctToken, result.mSbjctStrand, 0, result.mSbjctGenomeFrom, lgenome)
            if lgenome - result.mSbjctGenomeTo >= self.mMinGenomeSize:
                key = "%s:%s:%i:%i" % (result.mSbjctToken, result.mSbjctStrand, result.mSbjctGenomeTo, lgenome)                
                fragments[key] = ( result.mSbjctToken, result.mSbjctStrand, result.mSbjctGenomeTo, lgenome, lgenome )

        if not fragments: return results

        new_genome_files = []
            
        first = True
        single = len(genome) == 1
        for key, data in fragments.items():

            sbjct_token, sbjct_strand, xfrom, xto, lgenome = data
            ## extract sequence to be read
            if Genomics.IsNegativeStrand( sbjct_strand ):
                xfrom, xto = lgenome - xto, lgenome - xfrom

            ## if single sequence genomes, keep them that way
            ## otherwise: combine them
            if single or first:
                if not first: os.close(new_outfile_genome)
                
                new_outfile_genome, new_filename_genome = tempfile.mkstemp()                
                new_genome_files.append( new_filename_genome )
                first = False
                                    
            os.write(new_outfile_genome, ">%s\n%s\n" % (key, genome[sbjct_token][xfrom:xto]) )

            if self.mLogLevel >= 2:
                sys.stdout.write( "# adding fragment to %s: %s for %s:%s-%s.\n" % (new_filename_genome, key, sbjct_token, xfrom, xto) )
            
        os.close(new_outfile_genome)

        ## run recursive predictions
        new_results = Prediction.Predictions()
        for new_filename_genome in new_genome_files:
            new_results.combine( self.runRecursive( filename_peptides,
                                                    new_filename_genome,
                                                    options ) )
            

        ## change coordinates of predictions so that they correspond to the original
        ## submitted contigs.
        for result in new_results:
            sbjct_token, sbjct_strand, sbjct_from, sbjct_to, lgenome = fragments[r.mSbjctToken]
            # shift aligned region for fragment
            # this is strand dependent.
            r.shiftGenomicRegion( sbjct_from, lgenome - sbjct_to )
            ## set sbjct token to original
            r.mSbjctToken = sbjct_token
            ## sort out strand
            ## if match is on negative strand
            if Genomics.IsNegativeStrand(r.mSbjctStrand):
                ## switch coordinates
                sbjct_from, sbjct_to = lgenome - sbjct_to, lgenome-sbjct_from
                if Genomics.IsNegativeStrand(sbjct_strand):
                    sbjct_strand = "+"
                else:
                    sbjct_strand = "-"
            
            r.mSbjctStrand = sbjct_strand

        ## clean up
        os.remove( new_filename_genome )

        results.combine( new_results )
        return results
    
    def __call__( self,
                  filename_peptides,
                  filename_genome,
                  options = None):
        """call runRecursive."""

        return self.runRecursive( filename_peptides,
                                  filename_genome,
                                  options )
                

def getPredictor( options, parser_options ):
    """setup predictor."""

    ## sort out parsers:
    output_options = None
    if options.pretty:
        parser_options = {}
        prediction_parser = PredictionParserEcho()
        if options.method == "genewise":
            output_options = "-pretty"
        elif options.method == "genewise-hmm":
            output_options = "-pretty"
        elif options.method == "exonerate":
            output_options = "--showalignment"
    else:
        if options.method == "genewise":
            prediction_parser = PredictionParserGenewise( **parser_options )
        elif options.method == "genewise-hmm":
            if 'filename_peptides' in parser_options:
                del parser_options['filename_peptides']
            prediction_parser = PredictionParserGenewiseHMM( **parser_options )
        elif options.method == "exonerate":
            prediction_parser = PredictionParserExonerate( **parser_options )                                    
        
    if options.method == "genewise":
        predictor = PredictorGenewise( parser = prediction_parser )
    elif options.method == "genewise-hmm":
        predictor = PredictorGenewiseHMM( parser = prediction_parser )
    elif options.method == "exonerate":
        predictor = PredictorExonerate( parser = prediction_parser )
    else:
        raise "Unknown method %s" % (options.method)

    if output_options:
        predictor.setOutputOptions( output_options )

    if options.dry_run:
        predictor.setDryRun( options.dry_run )

    predictor.setLogLevel( options.loglevel )

    if options.use_cluster:
        cluster = Cluster.ClusterSGE( options )
        predictor.setCluster( cluster )

        tempfile.tempdir = "."
        
    if options.recursive:
        predictor = RecursivePredictor( predictor )
        predictor.setLogLevel( options.loglevel )
        
    if options.downstream_queries or options.upstream_queries:
        predictor = TandemPredictor( predictor,
                                     upstream_queries = options.upstream_queries,
                                     downstream_queries = options.downstream_queries )
        predictor.setLogLevel( options.loglevel )
        
    return predictor

def runLocally( filename_peptides, filename_genome, options, parser_options ):
    """run jobs locally."""
    
    result = Prediction.Predictions()
    
    if options.genome_chunk_size:
        infile = open(filename_genome, "r")
        filename_genomes = IndexedFasta.splitFasta( infile,
                                                 options.genome_chunk_size,
                                                 dir=os.curdir )
        infile.close()
        
        if options.loglevel >= 1:
            options.stdlog.write("# starting %i jobs.\n" % len(filename_genomes) )

        x = 0
        for filename_genome in filename_genomes:
            
            x += 1
            
            if options.loglevel >= 1:
                options.stdlog.write("# running job %i on %s.\n" % (x, filename_genome ) )
            
            predictor = getPredictor( options, parser_options )

            r = predictor( filename_peptides,
                           filename_genome,
                           options = options.options )

            if options.loglevel >= 2:
                options.stdlog.write("# %s\n" % (str(r)))
            
            result.combine( r )
            
    else:
        predictor = getPredictor( options, parser_options )
        
        result = predictor( filename_peptides,
                            filename_genome,
                            options = options.options )
    
    return result

def runOnCluster( filename_peptides, filename_genome, options, parser_options ):
    """run jobs on cluster.

    Warning: does not work perfectly - there are some problems
    with the parsers which surface when called from inside a makefile,
    but (strangely) not when called from the shell directly.
    """
    all_results = Prediction.Predictions()

    def save_result( request, result ):
        if result:
            all_results.combine( result )            

    def report_error( request, exc_info ):
        options.stdlog.write( "# exception occured in request #%s: %s\n" % \
                              (request.requestID, exc_info[1]) )
        options.stdlog.flush()

    def predict( predictor, filename_peptides, filename_genome, options ):
        result = predictor( filename_peptides, filename_genome, options = options)
        return result
    
    if options.genome_chunk_size:
        infile = open(filename_genome, "r")
        filename_genomes = IndexedFasta.splitFasta( infile,
                                                    options.genome_chunk_size,
                                                    dir=os.curdir )
        infile.close()
        
        if options.loglevel >= 1:
            options.stdlog.write("# starting %i parallel jobs with a pool size of %i.\n" %\
                                 ( len(filename_genomes), options.cluster_num_jobs ) )
            options.stdlog.flush()
            
        ## run jobs in parallel

        import threadpool
        pool = threadpool.ThreadPool( options.cluster_num_jobs )

        for filename_genome in filename_genomes:
            predictor = getPredictor( options, parser_options )
            request = threadpool.WorkRequest( predict,
                                              (predictor, filename_peptides, filename_genome, options.options),
                                              (),
                                              None,
                                              save_result,
                                              report_error )
            
            pool.putRequest( request )

        pool.wait()
        
    return all_results

def predictFromGFF( options ):
    """run predictions from a gff file that associates queries with genomic regions.
    """
    
    if not options.genome_file or not options.filename_peptides or options.filename_peptides == "-":
        raise "the --from-pairs option requires both a genome and an indexed set of queries."""

    fasta = IndexedFasta.IndexedFasta( options.genome_file )
    p, e = os.path.splitext(options.filename_peptides)
    if e in (".fasta" , ".fa" ):
        peptides = Genomics.ReadPeptideSequences( open(options.filename_peptides, "r") )
    else:
        peptides = IndexedFasta.IndexedFasta( options.filename_peptides )
        
    ninput, noutput, nskipped = 0, 0, 0

    if options.from_gff == "-":
        infile = sys.stdin
    else:
        infile = open( options.from_gff, "r" )
        
    predictor = getPredictor( options, parser_options )

    options.stdout.write( "%s\n" % Prediction.Prediction().getHeader() )

    prediction_id = 0
    for gffs in GTF.joined_iterator( GTF.iterator( infile ) ): 

        ninput += 1

        gffs.sort( lambda x,y: cmp(x.start, y.start) )

        g = gffs[0]
        contig, strand = g.name, g.strand
        if "Query" in g:
            query = g["Query"]
        else:
            query = g.info

        if "Id" in g:
            id = g["Id"]
        else:
            prediction_id += 1
            id = prediction_id

        if contig not in fasta:
            if options.loglevel >= 1:
                options.stdlog.write( "# contig %s not in genome - query %s skipped\n" % (contig, query))
                options.stdlog.flush()
            nskipped += 1
            continue

        if query not in peptides:
            if options.loglevel >= 1:
                options.stdlog.write( "# query %s not in peptides - query %s skipped\n" % (query, query))
                options.stdlog.flush()
            nskipped += 1
            continue

        start, end = gffs[0].start, gffs[-1].end

        lcontig = fasta.getLength( contig )

        if end > lcontig:
            raise "region for query %s out of bounds: %s:%i > %i" % (query, contig, end, lcontig)

        if options.add_flank:
            start = max( 0, start - options.add_flank )
            end = min( lcontig, end + options.add_flank )

        if Genomics.IsNegativeStrand( strand):
            start, end = lcontig - end, lcontig - start
            
        if options.loglevel >= 2:
            options.stdlog.write( "# predicting query %s in region %s:%s:%i:%i\n" %\
                                      (query, contig, strand, start, end))
            options.stdlog.flush()

        result = predictor.predictFromRange( 
            peptides[query],
            contig,
            strand,
            start, 
            end,
            fasta,
            options = options.options )
        
        noutput += 1

        for r in result:
            r.mPredictionId = id
            r.mQueryToken = query
            options.stdout.write( "%s\n" % str(r))
            
        options.stdout.flush()

    if options.from_gff != "-": infile.close()

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

def predictFromPairs( options ):
    """run predictions from two fasta files, one with peptide sequences and the
    other with genomic sequences. 

    Sequences are associated via the identifier.

    The cds file should be indexd.
    """
    
    if not options.genome_file or not options.filename_peptides or options.filename_peptides == "-":
        raise "the --from-pairs option requires both a genome and an indexed set of queries."""


    try:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
    except KeyError:
        fasta = Genomics.ReadGenomicSequences( open(options.genome_file, "r"), do_reverse = False )

    try:
        peptides = IndexedFasta.IndexedFasta( options.filename_peptides )
        keys = peptides.getContigs()
    except KeyError:
        peptides = Genomics.ReadPeptideSequences( open(options.filename_peptides, "r") )
        keys = peptides.keys()

    ninput, noutput, nskipped = 0, 0, 0

    predictor = getPredictor( options, parser_options )

    options.stdout.write( "%s\n" % Prediction.Prediction().getHeader() )

    for query in keys:

        ninput += 1
        peptide_sequence = peptides[query]

        if query not in fasta:
            E.warn( "query %s not cds database: skipped" % (query,))
            nskipped += 1
            continue

        genomic_sequence = fasta[query]

        E.debug( "predicting query %s against sequence of %i" % (query, len(genomic_sequence)))

        result = predictor.predictFromSequences( 
            peptide_sequence,
            genomic_sequence, 
            options = options.options )
        
        noutput += 1

        for r in result:
            r.mPredictionId = query
            r.mQueryToken = query
            options.stdout.write( "%s\n" % str(r))
            
        options.stdout.flush()

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: Predictor.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-p", "--filename-peptides", dest="filename_peptides", type ="string",
                       help="filename with peptide sequences. Use - for stdin.")

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("genewise", "exonerate", "genewise-hmm"),
                      help="""gene prediction method to use. If genewise-hmm is chosen, the peptide sequence file
                      is assumed to contain a profile."""  )
    
    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome pattern."  )
    
    parser.add_option("-r", "--range", dest="range", type="string",
                      help="range to scan."  )

    parser.add_option( "--from-gff", dest="from_gff", type="string",
                       help="""read a list of query identifiers and genomic locations from a tab-separated file (- for stdin) and run a 
predictor for each. 
""" )

    parser.add_option( "--from-pairs", dest="from_pairs", action="store_true",
                       help="""predict pairs of peptide/cds sequences from the two fasta files. The identifiers must match""")

    parser.add_option("-R", "--filename-ranges", dest="filename_ranges", type="string",
                      help="filename with genomic ranges to scan a set of queries against."  )

    parser.add_option("-o", "--options", dest="options", type="string",
                       help="options to pass on to predictor.")

    parser.add_option("-i", "--pretty", dest="pretty", action="store_true",
                      help="pretty print results directly.")
    
    parser.add_option( "--dry-run", dest="dry_run", action="store_true",
                      help="prepare files, but do not execute prediction program.")

    parser.add_option( "--forward-coordinates", dest="forward_coordinates", action="store_true",
                      help="return result in forward coordinates.")

    parser.add_option( "--both-strands", dest="both_strands", action="store_true",
                      help="analyze both strands.")

    parser.add_option( "--use-fasta-coordinates", dest="use_fasta_coordinates", action="store_true",
                       help="""use fasta coordinates. The format of the description line should thus be
                       >id contig:strand:from:to:lcontig in 0-based reverse coordinates.
                       """ )

    parser.add_option( "--recursive", dest="recursive", action="store_true",
                      help="analyze genomic stretches recursively.")

    parser.add_option( "--downstream-queries", dest="downstream_queries", type="string", action="append",
                      help="downstream queries, triggered for each match.")
    
    parser.add_option( "--upstream-queries", dest="upstream_queries", type="string", action="append",
                      help="upstream queries, triggered for each match.")

    parser.add_option( "--genome-chunk-size", dest="genome_chunk_size", type="int",
                      help="chunk size of genomic files.")
    
    parser.add_option( "--input-coordinate-format", dest="input_coordinate_format", type="choice",
                       choices=( "zero-both", "zero-forward" ),
                       help="coordinate format." )

    parser.add_option( "--add-flank", dest="add_flank", type="int",
                      help="add a flanking region to each region to be queried.")
    
    parser.set_defaults(
        method = "exonerate",
        options   = "",
        genome_file = "genome",
        filename_peptides = "-",
        filename_ranges = None,
        filename_profile = None,
        range = None,
        pretty = False,
        dry_run = False,
        forward_coordinates = False,
        both_strands = False,
        use_fasta_coordinates = False,
        recursive = False,
        genome_chunk_size = 0,
        downstream_queries = [],
        upstream_queries = [] ,
        input_coordinate_format= "zero-both",
        add_flank = 0,
        from_pairs = False,
        from_gff = None,
        )

    (options, args) = E.Start( parser,
                               add_pipe_options = True,
                               add_cluster_options = True )

    parser_options = {}

    ##########################################################################
    ##########################################################################
    ##########################################################################    
    ## setup input files
    ##########################################################################
    ## translation table between internal contig names
    ## and original locations:
    map_key2contig = {}    

    if options.from_gff:

        predictFromGFF( options )
        E.Stop()
        sys.exit(0)

    elif options.from_pairs:
        
        predictFromPairs( options )
        E.Stop()
        sys.exit(0)
        
    if options.filename_ranges:

        fasta = IndexedFasta.IndexedFasta( options.genome_file )
        
        if options.filename_ranges == "-":
            infile = sys.stdin
        else:
            infile = open(options.filename_ranges)
            
        genomic_sequences = {}
        
        outfile, filename_genome = tempfile.mkstemp()
       
        for line in infile:
            if line[0] == "#": continue

            sbjct_token, sbjct_strand, sbjct_from, sbjct_to = line[:-1].split(":")

            sbjct_from, sbjct_to = int(sbjct_from), int(sbjct_to)

            key = "%s:%s:%s:%s" % (sbjct_token, sbjct_strand, sbjct_from, sbjct_to)
            
            genomic_sequence = fasta.getSequence( sbjct_token, sbjct_strand,
                                                  sbjct_from, sbjct_to )
            
            genomic_sequences[key] = genomic_sequence

            map_key2contig[key] = (sbjct_token, sbjct_strand, sbjct_from, sbjct_to )
        
            os.write(outfile, ">%s\n%s\n" % (key, genomic_sequence))

        os.close(outfile)

        if options.filename_ranges == "-":
            infile.close()

        parser_options['genomes'] = genomic_sequences

    elif options.range:
        
        fasta = IndexedFasta.IndexedFasta( options.genome_file )

        sbjct_token, sbjct_strand, sbjct_from, sbjct_to = Genomics.String2Location( options.range )

        ## convert to my coordinate format
        converter = IndexedFasta.getConverter( options.input_coordinate_format )        
        sbjct_from, sbjct_to = converter( sbjct_from, sbjct_to,
                                          Genomics.IsPositiveStrand(sbjct_strand),
                                          fasta.getLength( sbjct_token ) )

        genomic_sequence = fasta.getSequence( sbjct_token, sbjct_strand,
                                              sbjct_from, sbjct_to )

        key = "%s:%s:%s:%s" % (sbjct_token, sbjct_strand, sbjct_from, sbjct_to)
        
        map_key2contig[key] = (sbjct_token, sbjct_strand, sbjct_from, sbjct_to )
        
        genomic_sequences = { key : genomic_sequence }
        
        outfile, filename_genome = tempfile.mkstemp()
        os.write(outfile, ">%s\n%s\n" % (key, genomic_sequence))
        os.close(outfile)
        parser_options['genomes'] = genomic_sequences
        
    else:
        filename_genome = options.genome_file
        parser_options['filename_genome'] = filename_genome
        
    peptide_sequences = {}
    if options.filename_peptides == "-":
        ## save peptides to temporary file
        outfile, filename_peptides = tempfile.mkstemp()
        for line in sys.stdin:
            os.write(outfile, line)
        os.close(outfile)
        parser_options['filename_peptides'] = filename_peptides
    else:
        filename_peptides = options.filename_peptides
        parser_options['filename_peptides'] = options.filename_peptides

    ##########################################################################
    ##########################################################################
    ##########################################################################    
    ## setup cluster run by splitting peptide and or genomic files
    ##########################################################################
    if options.use_cluster:
        result = runOnCluster( filename_peptides, filename_genome, options, parser_options )
    else:
        result = runLocally( filename_peptides, filename_genome, options, parser_options )        

    E.info( "obtained %i results" % len(result))

    ##########################################################################
    ##########################################################################
    ##########################################################################    
    ## post-processing
    ##########################################################################
    if options.range:
        if not options.dry_run:
            os.remove(filename_genome)
        for r in result:
            sbjct_token, sbjct_strand, sbjct_from, sbjct_to = map_key2contig[r.mSbjctToken]
            lgenome = fasta.getLength( sbjct_token )
            # shift aligned region for fragment
            # this is strand dependent.
            r.shiftGenomicRegion( sbjct_from, lgenome - sbjct_to )
            ## set sbjct token to original
            r.mSbjctToken = sbjct_token
            ## sort out strand
            ## if match is on negative strand
            if r.mSbjctStrand in ("-", "0", -1, "-1"):
                ## switch coordinates
                sbjct_from, sbjct_to = lgenome - sbjct_to, lgenome-sbjct_from
                if sbjct_strand in ("-", "0", -1, "-1"):
                    sbjct_strand = "+"
                else:
                    sbjct_strand = "-"
            
            r.mSbjctStrand = sbjct_strand

    elif options.use_fasta_coordinates:
        ## get coordinates from fasta file.
        ## the contig length should be supplied in the file
        ## for this to work.
        coordinates = {}
        infile = open(filename_genome,"r")
        for line in infile:
            if line[0] == ">":
                tokens = re.split("\s+", line[1:-1] )
                if len(tokens) == 1:
                    coord = tokens[0]
                    id = coord
                else:
                    id, coord = tokens[:2]

                try:
                    contig, strand, xfrom, xto, lcontig = coord.split(":")
                except ValueError:
                    E.warn( "skipped: can not parse coordinates for %s = %s" % (id, coord))
                    continue

                coordinates[id] = ( contig, strand, int(xfrom), int(xto), int(lcontig) )
                
        infile.close()
        
        for r in result:
            sbjct_token, sbjct_strand, sbjct_from, sbjct_to, lcontig = coordinates[r.mSbjctToken]
            
            # shift aligned region for fragment
            # this is strand dependent.
            r.shiftGenomicRegion( sbjct_from, lcontig - sbjct_to )
            ## set sbjct token to original
            r.mSbjctToken = sbjct_token
            
            ## sort out strand
            ## if match is on negative strand
            if r.mSbjctStrand in ("-", "0", -1, "-1"):
                ## switch coordinates
                if sbjct_strand in ("-", "0", -1, "-1"):
                    sbjct_strand = "+"
                else:
                    sbjct_strand = "-"
            
            r.mSbjctStrand = sbjct_strand

            ## convert to forward coordinates, if so desired.
            if options.forward_coordinates:
                r.mSbjctGenomeFrom, r.mSbjctGenomeTo = lcontig - r.mSbjctGenomeTo, lcontig - r.mSbjctGenomeFrom

        ## do not convert again (lcontig from fasta object does not reflect
        ## true contig size).
                
        options.forward_coordinates = False

    if options.forward_coordinates:
        for r in result:
            lgenome = fasta.getLength( sbjct_token )
            if r.mSbjctStrand in ("-", "0", -1, "-1"):
                r.mSbjctGenomeFrom, r.mSbjctGenomeTo = lgenome - r.mSbjctGenomeTo, lgenome - r.mSbjctGenomeFrom

    options.stdout.write( "%s\n" % Prediction.Prediction().getHeader() )
    x = 0
    for r in result:
        x += 1
        r.mPredictionId = x
        options.stdout.write( "%s\n" % str(r))

    ## clean up
    if options.filename_peptides == "-" and not options.dry_run:
        os.remove( filename_peptides)

    E.Stop()




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Intervals.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
def combine( intervals ):
    """combine intervals.

    Overlapping intervals are concatenated into larger intervals.
    """
    if not intervals:
        return []

    new_intervals = []
    
    intervals.sort()
    first_from, last_to = intervals[0]
    
    for this_from, this_to in intervals[1:]:
        if this_from > last_to:
            new_intervals.append( (first_from, last_to ) )
            first_from, last_to = this_from, this_to
            continue

        if last_to < this_to:
            last_to = this_to
    
    new_intervals.append( ( first_from, last_to ))

    return new_intervals

#----------------------------------------------------------------
def prune( intervals, first = None, last = None):
    """truncates all intervals that are extending beyond first or last.

    Empty intervals are deleted.
    """
    new_intervals = []
    for start, end in intervals:
        if end <= first or start >= end: continue
        new_intervals.append( ( max( first, start), min( end, last) ) )
    return new_intervals

#----------------------------------------------------------------
def complement( intervals, first = None, last = None):
    """complement a list of intervals with intervals not in list.
    """
    
    if len(intervals) == 0:
        if first != None and last != None:
            return [(first,last)]
        else:
            return []
        
    new_intervals = []
    
    intervals.sort()
    last_from, last_to = intervals[0]

    if first != None and first < last_from:
        new_intervals.append( (first, last_from) )

    for this_from, this_to in intervals:
        if this_from > last_to:
            new_intervals.append( (last_to, this_from ) )            
            
        last_from = this_from
        last_to = max(last_to, this_to)

    if last and last > last_to:
        new_intervals.append( (last_to, last))

    return new_intervals

#----------------------------------------------------------------
def addComplementIntervals( intervals, first = None, last = None):
    """complement a list of intervals with intervals not
    in list and return both.
    """

    return intervals + complementIntervals( intervals, first, last)

#----------------------------------------
def combineAtDistance( intervals, min_distance ):
    """combine a list intervals and merge those that are less than a certain
    distance apart.
    """
    
    if not intervals: return []

    # merge overlapping intervals
    intervals = combine( intervals )

    new_intervals = []
    intervals.sort()
    
    first_start, last_end = intervals[0]
    
    for this_start, this_end in intervals[1:]:
        
        if this_start - last_end >= min_distance:
            new_intervals.append( (first_start, last_end ) )
            first_start = this_start

        last_end = this_end
    
    new_intervals.append( ( first_start, last_end ))

    return new_intervals

#----------------------------------------
def DeleteSmallIntervals( intervals, min_length ):
    """combine a list of non-overlapping intervals,
    and delete those that are too small.
    """
    
    if not intervals: return []

    new_intervals = []

    for this_from, this_to in intervals:
        if (this_to - this_from + 1) >= min_length:
            new_intervals.append( (this_from, this_to ) )
            
    return new_intervals

#----------------------------------------------------------------
def getIntersections( intervals ):
    """combine intervals.

    Overlapping intervals are reduced to their intersection.
    """
    if not intervals: return []

    intervals.sort()
    max_to = intervals[0][1]

    all_sections = []
    sections = [ intervals[0][0], intervals[0][1] ]
    for this_from, this_to in intervals[1:]:

        # no overlap: write everything and reset
        if this_from > max_to:
            all_sections.append(sections)
            max_to = this_to
            sections = []
            
        max_to = max(max_to, this_to)
        
        sections.append( this_from )
        sections.append( this_to )
        
    all_sections.append( sections )

    new_intervals = []
    for sections in all_sections:
        sections.sort()
        last_x = sections[0]
        for x in sections[1:]:
            if last_x == x: continue
            new_intervals.append( (last_x, x) )
            last_x = x
            
    return new_intervals



#----------------------------------------------------------------
def RemoveIntervalsContained( intervals ):
    """
    remove intervals that are fully contained in another.

    [(10, 100), (20, 50), (70, 120), (130, 200), (10, 50), (140, 210), (150, 200)]

    results:
    
    [(10, 100), (70, 120), (130, 200), (140, 210)]   
    """
    if not intervals:
        return []

    new_intervals = []
    
    intervals.sort()
    last_from, last_to = intervals[0]
    
    for this_from, this_to in intervals[1:]:
        # this is larger:
        if this_from <= last_from and this_to >= last_to:
            last_from,last_to = this_from,this_to
            continue

        # last is larger
        if last_from <= this_from and last_to >= this_to:
            continue

        # no complete overlap
        new_intervals.append( (last_from, last_to ) )

        last_from,last_to = this_from,this_to        
        
    new_intervals.append( ( last_from, last_to ))
    
    return new_intervals

#----------------------------------------------------------------
def RemoveIntervalsSpanning( intervals ):
    """remove intervals that are full covering
    another, i.e. always keep the smallest.

    [(10, 100), (20, 50), (70, 120), (40,80), (130, 200), (10, 50), (140, 210), (150, 200)]

    result:
    
    [(20, 50), (40, 80), (70, 120), (150, 200)]
    """
    
    if not intervals: return []
    
    intervals.sort()
    
    last_intervals = intervals
        
    while 1:

        new_intervals = []
    
        last_from, last_to = last_intervals[0]
    
        for this_from, this_to in last_intervals[1:]:
            # print last_from, last_to, this_from, this_to
            # this is larger:
            if this_from <= last_from and this_to >= last_to:
                continue
            
            # last is larger:
            if last_from <= this_from and last_to >= this_to:
                last_from,last_to = this_from,this_to
                continue

            # no complete overlap
            new_intervals.append( (last_from, last_to ) )
            last_from,last_to = this_from,this_to        
        
        new_intervals.append( ( last_from, last_to ))

        if len(last_intervals) == len(new_intervals): break

        last_intervals = new_intervals
    
    return new_intervals

#----------------------------------------------------------------
def ShortenIntervalsOverlap( intervals, to_remove ):
    """shorten intervals, so that there is no
    overlap with another set of intervals.

    assumption: intervals are not overlapping
    
    """
    if not intervals:
        return []

    if not to_remove:
        return interalls
    
    new_intervals = []
    
    intervals.sort()
    to_remove.sort()

    current_to_remove = 0
    
    for this_from, this_to in intervals:

        for remove_from, remove_to in to_remove:
            #print this_from, this_to, remove_from, remove_to
            if remove_to < this_from: continue
            if remove_from > this_to: continue

            if remove_from <= this_from and remove_to >= this_to:
                this_from = remove_to
                break

            if this_from < remove_from:
                new_intervals.append( (this_from, remove_from) )
                # print "adding", this_from, remove_from
            
            this_from = max(this_from, remove_to)

            if this_to < this_from: break
            
        if this_to > this_from:
            # print "adding", this_from, this_to
            new_intervals.append( (this_from, this_to) )            

    return new_intervals

#----------------------------------------------------------------
def joined_iterator( intervals1, intervals2 ):
    """iterate over the combination of two intervals.

    returns the truncated intervals delineating the
    ranges of overlap between intervals1 and intervals2.
    """

    intervals1.sort()
    intervals2.sort()
    
    x, y = 0, 0
    while x < len(intervals1) and y < len(intervals2):

        xfrom, xto = intervals1[x]
        yfrom, yto = intervals2[y]

        if xto <= yfrom:
            x += 1
        elif yto <= xfrom:
            y += 1
        else:
            mto = min( xto, yto )
            mfrom = max( xfrom, yfrom )

            yield (mfrom, mto)

            if xto < yto:
                x += 1
            elif yto < xto:
                y += 1
            else:
                x += 1
                y += 1

#----------------------------------------------------------------
def intersect( intervals1, intervals2 ):
    """intersect two interval sets.

    Return a set of intervals that is spanned by intervals in
    both sets. Returns the union of the two intervals.
    """
    if not intervals1 or not intervals2: return []
    
    result = []
    for start, end in joined_iterator( intervals1, intervals2 ):

        overlap = end - start
        if overlap: result.append( (start, end) )

    return result

#----------------------------------------------------------------
def getLength( intervals ):
    '''return sum of intervals.'''
    return sum([x[1]-x[0] for x in intervals])

#----------------------------------------------------------------
def truncate( intervals1, intervals2 ):
    """truncate intervals in intervals1 by intervals2

    Example: truncate( [(0,5)], [(0,3)] ) = [(3,5)]
    """
    if not intervals1: return []
    if not intervals2: return intervals1
    
    intervals2 = combine(intervals2)

    result = []
    intersection = intersect( intervals1, intervals2 )

    if not intersection: return intervals1

    y = 0
    for start, end in intervals1:
        
        # retrieve all overlapping in intersection
        ovl = []
        while y < len(intersection):
            mask_start, mask_end = intersection[y]            
            if mask_start >= end: break
            if mask_start >= start: ovl.append( (mask_start, mask_end) )
            y += 1
            
        # build truncated segments
        last = start
        for mask_start, mask_end in ovl:
            if mask_start - last > 0:
                result.append( (last, mask_start) )
            last = mask_end
        if end - last > 0:
            result.append( (last, end) )

    return result

#----------------------------------------------------------------
def calculateOverlap( intervals1, intervals2 ):
    """calculate overlap between intervals.

    The intervals within each set should not be overlapping.
    """

    if not intervals1 or not intervals2: return 0

    overlap = 0
    for start, end in joined_iterator( intervals1, intervals2 ):
        overlap += end - start

    return overlap
        
#----------------------------------------------------------------
def fromArray( a ):
    """get intervals from a binary array."""
    
    if len(a) == 0: return []

    was_a = a[0]
    if was_a: start = 0
    
    intervals = []
    for x in range(1, len(a)):
        is_a = a[x]
        if is_a and not was_a:
            start = x
            was_a = is_a
        elif not is_a and was_a:
            if start != None: intervals.append( (start, x) )
            was_a = is_a
    if is_a: intervals.append( (start, len(a) ) )
        
    return intervals



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
bed2table.py - annotate intervals
=================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script takes a bed-formatted file as input and annotates each interval.
Possible annotators are (see option '--counter'):

overlap
    compute overlap with intervals in other bed file. If the other bed 
    file contains tracks, the overlap is computed per track.

peaks
    compute peak location in intervals. Requires one or more bam-files. This
    counter can also count within an secondary set of bam-files (--control-bam-file)
    and add this to the output.

composition-na
    compute nucleotide frequencies in intervals.

composition-cpg
    compute CpG densities and nucleotide frequencies in intervals. 

Usage
-----

Example::

   python bed2table.py --help

Type::

   python bed2table.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math
import time
import tempfile
import subprocess
import types
import bisect
import array
import collections
import CGAT.GTF as GTF
import CGAT.Bed as Bed
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Stats as Stats
import CGAT.SequenceProperties as SequenceProperties
import CGAT.Genomics as Genomics
import CGAT.Intervals as Intervals
import numpy
import CGAT.IndexedGenome as IndexedGenome
import pysam

class Counter( object ):
    
    def __init__(self, fasta = None, *args, **kwargs ):
        self.fasta = fasta

class CounterOverlap( Counter ):
    '''count overlap for each interval in tracks.'''

    def __init__(self, filename, *args, **kwargs ):
        
        assert filename != None, "please supply filename for CounterOverlap"

        Counter.__init__(self, *args, **kwargs )

        self.filename = filename

        E.info( "reading intervals from %s" % self.filename )

        self.index = Bed.readAndIndex( open( self.filename, "r"),
                                       per_track = True )
        
        E.info( "read intervals for %s tracks" % len(self.index) )

        self.tracks = self.index.keys()
        self.headers = []
        for track in self.tracks:
            self.headers.extend( ["%s_nover" % track, "%s_bases" % track] )
            
    
    def update( self, bed ):
        '''update internal counts.'''

        results = []
        for track in self.tracks:
            try:
                overlaps = [ (x[0],x[1]) for x in self.index[track][bed.contig].find( bed.start, bed.end ) ]
            except KeyError:
                overlaps = []

            results.append( (len(overlaps), 
                             Intervals.calculateOverlap( [(bed.start, bed.end),],
                                                         Intervals.combine( overlaps ) ) ) )

        self.data = results

    def __str__(self):
        '''output overlap of interval in *bed*'''
                        
        r = []
        for track, result in zip(self.tracks, self.data):
            r.append( "\t".join( (str(result[0]), str(result[1])) ) )

        return "\t".join(r)

##-----------------------------------------------------------------------------------
CounterPeaksResult = collections.namedtuple( "CounterPeaksResult", ("length nreads avgval peakval npeaks peakcenter" ) )
class CounterPeaks(Counter):
    '''compute number of extent of peaks in an interval.'''

    headers = None

    def __init__(self, 
                 bamfiles, 
                 offsets, 
                 control_bamfiles, 
                 control_offsets, *args, **kwargs ):
        Counter.__init__(self, *args, **kwargs )
        if not bamfiles: raise ValueError("supply --bam-file options for readcoverage")

        assert len(offsets) == 0 or len(bamfiles) == len(offsets), "number of bamfiles not the same as number of offsets"
        assert len(control_offsets) == 0 or len(control_bamfiles) == len(control_offsets), "number of control bamfiles not the same as number of offsets"

        self.bamfiles = bamfiles
        self.offsets = offsets
        self.control_bamfiles = control_bamfiles
        self.control_offsets = control_offsets

        self.headers = list(CounterPeaksResult._fields)
        if self.control_bamfiles:
            self.headers.extend( ["control_%s" % x for x in CounterPeaksResult._fields] )

    def count( self, bed, bamfiles, offsets ):
        '''count reads in bed interval.'''

        contig, start, end = bed.contig, bed.start, bed.end

        length = end - start
        try:
            counts = numpy.zeros( length )
        except ValueError, msg:
            raise ValueError( "Error negative length obtained: message=%s contig=%s, start=%s, end=%s" %(msg, contig, start, end))
        nreads = 0

        if offsets:
            # if offsets are given, shift tags. 
            for samfile, offset in zip(bamfiles,offsets):

                if offset == 0: E.warn("0 offset will result in no data!" )
                shift = offset / 2
                # for peak counting I follow the MACS protocoll,
                # see the function def __tags_call_peak in PeakDetect.py
                # In words
                # Only take the start of reads (taking into account the strand)
                # add d/2=offset to each side of peak and start accumulate counts.
                # for counting, extend reads by offset
                # on + strand shift tags upstream
                # i.e. look at the downstream window
                xstart, xend = max(0, start - shift), max(0, end + shift)

                for read in samfile.fetch( contig, xstart, xend ):
                    nreads += 1
                    pos = read.pos
                    # some reads are assigned to a contig and position, but
                    # are flagged as unmapped - these might not have an alen attribute.
                    if read.is_unmapped: continue

                    if read.is_reverse:
#                        rstart = read.pos + read.alen - offset
                       # offset = 2 * shift
                        try:
                            rstart = read.pos + read.alen - offset
                        except TypeError, msg:
                            raise TypeError("Error message =", msg, "read.pos =", read.pos, "read.alen =", read.alen, "offset =", offset, "query name =", read.qname, "length of read =", read.rlen)
                    else: 
                        rstart = read.pos + shift

                    rend = rstart + shift
                    rstart = max( 0, rstart - start )
                    rend = min( length, rend - start )
                    counts[ rstart:rend ] += 1

        else:
            for samfile in bamfiles:
                for read in samfile.fetch( contig, start, end ):
                    nreads += 1
                    rstart = max( 0, read.pos - start )
                    rend = min( length, read.pos - start + read.rlen ) 
                    counts[ rstart:rend ] += 1
                    
        length = end - start            
        avgval = numpy.mean( counts )
        peakval = max(counts)

        # set other peak parameters
        peaks = numpy.array( range(0,length) )[ counts >= peakval ]
        npeaks = len( peaks )
        # peakcenter is median coordinate between peaks
        # such that it is a valid peak in the middle
        peakcenter = start + peaks[npeaks//2] 

        return CounterPeaksResult( length, nreads, avgval, peakval, npeaks, peakcenter )

    def update( self, bed ):
        '''count reads per position.
        
        If offsets are given, shift tags by offset / 2 and extend
        by offset / 2.
        '''

        self.result = self.count( bed, self.bamfiles, self.offsets )
        if self.control_bamfiles:
            self.control = self.count( bed, self.control_bamfiles, self.control_offsets )

    def __str__(self):
        if self.control_bamfiles:
            return "\t".join( map(str, self.result + self.control ) )
        else:
            return "\t".join( map(str, self.result ) )

##-----------------------------------------------------------------------------------
class CounterCompositionNucleotides(Counter):

    headers = SequenceProperties.SequencePropertiesNA().getHeaders() 

    def __init__(self, *args, **kwargs ):
        Counter.__init__(self, *args, **kwargs )
        self.result_class = SequenceProperties.SequencePropertiesNA
        
    def update(self, bed):
        s = self.fasta.getSequence( bed.contig, "+", bed.start, bed.end)
        self.result = self.result_class()
        self.result.loadSequence( s )

    def __str__(self):
        return str(self.result)

##-----------------------------------------------------------------------------------
class CounterCompositionCpG(CounterCompositionNucleotides):
    '''compute CpG frequencies as well as nucleotide frequencies.

    Note that CpG density is calculated across the merged exons
    of a transcript. Thus, there might be difference between the CpG 
    on a genomic level and on the transrcipt level depending on how
    many genomic CpG are lost across an intron-exon boundary or how
    many transcript CpG are created by exon fusion.
    '''

    headers = SequenceProperties.SequencePropertiesCpg().getHeaders() 

    def __init__(self, *args, **kwargs ):
        CounterCompositionNucleotides.__init__(self, *args, **kwargs )
        self.result_class = SequenceProperties.SequencePropertiesCpg

    def update(self, bed):
        try:
            s = self.fasta.getSequence( bed.contig, "+", bed.start, bed.end+1)
            next_char = s[-1]
            s = s[:-1]
        except ValueError:
            s = self.fasta.getSequence( bed.contig, "+", bed.start, bed.end)
            next_char = None

        self.result = self.result_class()
        self.result.loadSequence( s, next_char = next_char )

##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: gtf2table.py 2888 2010-04-07 08:48:36Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )

    parser.add_option("-b", "--bam-file", dest="bam_files", type="string",
                      help="filename with read mapping information. Multiple files can be submitted in a comma-separated list [default=%default]."  )

    parser.add_option( "--control-bam-file", dest="control_bam_files", type="string",
                      help="filename with read mapping information for input/control. Multiple files can be submitted in a comma-separated list [default=%default]."  )

    parser.add_option("-c", "--counter", dest="counters", type="choice", action="append",
                      choices=( "overlap", "peaks", "composition-na", "composition-cpg" ),
                      help="select counters to apply [default=%default]."  )

    parser.add_option("-o", "--offset", dest="offsets", type="int", action="append",
                      help="tag offsets for tag counting - supply as many as there are bam-files [default=%default]."  )

    parser.add_option( "--control-offset", dest="control_offsets", type="int", action="append",
                      help="control tag offsets for tag counting - supply as many as there are bam-files [default=%default]."  )

    parser.add_option("-a", "--all-fields", dest="all_fields", action = "store_true",
                      help="output all fields in original bed file, by default only the first 4 are output [default=%default]."  )

    parser.add_option("--bed-headers", dest="bed_headers", type="string",
                      help="supply ',' separated list of headers for bed component [default=%default]."  )

    parser.set_defaults(
        genome_file = None,
        counters = [],
        bam_files = None,
        offsets = [],
        control_bam_files = None,
        control_offsets = [],
        all_fields = False,
        bed_headers = "contig,start,end,name",
        )

    (options, args) = E.Start( parser )

    # get files
    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
    else:
        fasta = None

    if options.bam_files:
        bam_files = []
        for bamfile in options.bam_files.split(","):
            bam_files.append( pysam.Samfile(bamfile, "rb" ) )
    else:
        bam_files = None

    if options.control_bam_files:
        control_bam_files = []
        for bamfile in options.control_bam_files.split(","):
            control_bam_files.append( pysam.Samfile(bamfile, "rb" ) )
    else:
        control_bam_files = None

    counters = []

    for c in options.counters:
        if c == "overlap":
            counters.append( CounterOverlap( filename = options.filename_bed,
                                             fasta=fasta,
                                             options = options) )
        elif c == "peaks":
            counters.append( CounterPeaks( bam_files,
                                           options.offsets,
                                           control_bam_files,
                                           options.control_offsets,
                                           options = options ) )
        elif c == "composition-na":
            counters.append( CounterCompositionNucleotides( fasta=fasta,
                                                            options = options ))
        elif c == "composition-cpg":
            counters.append( CounterCompositionCpG( fasta=fasta,
                                                    options = options ) )


    options.stdout.write( "\t".join( [x.strip() for x in options.bed_headers.split(",") ] ) )
    for counter in counters: 
        options.stdout.write("\t%s" % "\t".join( counter.headers ) )
    options.stdout.write("\n")

    for bed in Bed.iterator(options.stdin):
        if options.all_fields:
            options.stdout.write( str(bed) )
        else:
            options.stdout.write( "\t".join( (bed.contig, 
                                              str(bed.start), 
                                              str(bed.end), 
                                              bed.fields[0]) ) )
        for counter in counters: 
            counter.update(bed)
            options.stdout.write("\t%s" % str(counter) )
        options.stdout.write("\n")

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Fastq.py - methods for dealing with fastq files
===============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''

from math import log
# see http://en.wikipedia.org/wiki/FASTQ_format
# ranges are conservative - they are open-ended
RANGES = {
    'sanger' : (33, 75) ,
    'illumina-1.8' : (33, 76) ,
    'solexa' : (59, 106) ,
    'phred64' : (64, 106) ,
    }

class Record:

    def __init__( self, identifier, seq, quals, format = None ):
        self.identifier, self.seq, self.quals, format = identifier, seq, quals, format
        self.format = None

    def __str__(self ):
        return "@%s\n%s\n+\n%s" % (self.identifier, self.seq, self.quals )

    def guessFormat( self ):
        '''return quality score format - 
        might return several if ambiguous.'''
        
        c = [ord(x) for x in self.quals ]
        mi, ma = min(c), max(c)
        r = []
        for format, v in RANGES.iteritems():
            m1, m2 = v
            if mi >= m1 and ma < m2: r.append( format )
        return r

    def trim( self, trim3, trim5 = 0 ):
        self.seq = self.seq[trim5:-trim3]
        self.quals = self.quals[trim5:-trim3]

    def toPhred( self ):
        '''return qualities as a list of phred-scores.'''
        assert self.format != None, "format needs to be set for conversion"
        if self.format == "sanger":
            return [ ord(x) - 33 for x in self.quals ]
        elif self.format == "solexa":
            # from -5 to 40 (i.e., can be negative)
            log10x = log(10.0) + .499
            return [ int(10.0 * log( 1.0 + 10 ** (ord(x) / 10.0), 10) / log10x ) for x in self.quals ]
        elif self.format == "phred64":
            return [ ord(x) - 64 for x in self.quals ]

    def fromPhred( self, quals, format ):
        '''set qualities from a list of phred-scores.'''
        self.format = format
        # -1 for color space fastq file
        assert len(quals) == len(self.seq) or len(quals) == len(self.seq) - 1
        if self.format == "sanger":
            self.quals = "".join( [ chr(33 + x) for x in quals ] )
        elif self.format == "solexa":
            log10x = log(10.0, 10) / 10.0
            q = [int(10.0 * (log( 10 ** (x * log10x) - 1.0, 10 ))) for x in quals ] 
            self.quals = "".join( [ chr(64 + x) for x in q ] )
        elif self.format == "phred64":
            self.quals = "".join( [ chr(64 + x) for x in quals ] )
        elif self.format == "integer":
            self.quals = " ".join( map( str, quals ) )

def iterate( infile ):
    '''iterate over contents of fastq file.'''

    while 1:
        line1 = infile.readline()
        if not line1: break
        if not line1.startswith('@'):
            raise ValueError( "parsing error: expected '@' in line %s" % line1 )
        line2 = infile.readline()
        line3 = infile.readline()
        if not line3.startswith('+'):
            raise ValueError( "parsing error: expected '+' in line %s" % line3 )
        line4 = infile.readline()
        # incomplete entry
        if not line4: raise ValueError( "incomplete entry for %s" % line1)
        
        yield Record( line1[1:-1], line2[:-1], line4[:-1])


def iterate_guess( infile, max_tries = 10000):
    '''iterate over contents of fastq file.

    guess quality format.
    '''
    
    quals = set( RANGES.keys() )
    cache = []
    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        quals.intersection_update( set(record.guessFormat()) )
        if len(quals) == 0:
            raise ValueError( "could not guess format - ranges incompatible." )
        if len(quals) == 1:
            break
        cache.append( record )

        if c > max_tries: 
            break

    if len(quals) == 1:
        ref_format = list(quals)[0]
    else:
        raise ValueError( "could not guess format - could be one of %s." % str(quals) )
            
    for r in cache:
        r.format = ref_format
        yield r

    for r in myiter:
        r.format = ref_format
        yield r

def iterate_convert( infile, format, max_tries = 10000, guess = None ):
    '''iterate over contents of fastq file.

    guess quality format and set it to new format.
    '''
    
    quals = set( RANGES.keys() )
    cache = []
    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        quals.intersection_update( set(record.guessFormat()) )

        if len(quals) == 0:
            raise ValueError( "could not guess format - ranges incompatible." )
        if len(quals) == 1:
            break
        cache.append( record )

        if c > max_tries: 
            break

    if len(quals) == 1:
        ref_format = list(quals)[0]
    elif guess in quals:
        ref_format = guess
    else:
        raise ValueError( "could not guess format - could be one of %s." % str(quals) )
            
    for r in cache:
        r.format = ref_format
        r.fromPhred( r.toPhred(), format )
        yield r

    for r in myiter:
        r.format = ref_format
        r.fromPhred( r.toPhred(), format )
        yield r
         

def guessFormat( infile, max_lines = 10000, raises = True ):
    '''guess format of FASTQ File.

    If *raises* , a ValueError is raised if there is not a single format.
    
    '''
    
    quals = set( RANGES.keys() )
    myiter = iterate(infile)
    for c, record in enumerate(myiter):
        quals.intersection_update( set(record.guessFormat()) )
        if len(quals) == 0:
            raise ValueError( "could not guess format - ranges incompatible." )
        if len(quals) == 1:
            break
        if c > max_lines: 
            break

    if len(quals) == 1:
        return list(quals)[0]
    elif raises == False:
        return quals
    else:
        raise ValueError( "could not guess format - could be one of %s." % str(quals) )


def getOffset( format, raises = True ):
    '''returns the ASCII offset for a certain format.

    If *raises* , a ValueError is raised if there is not a single offset.

    Otherwise, a minimum offset is returned.
    '''
    if type(format) in (set, list, tuple):
        offsets = set([ RANGES[f][0] for f in format ])
        if len(offsets) == 1:
            return list(offsets)[0]
        elif raises == False:
            return min(offsets)
        else:
            raise ValueError("inconsistent offsets for multiple formats: %s" % offsets)

    return RANGES[format][0]


